@@ -41,14 +41,13 @@ int set_essid(struct wrap_ndis_device *wnd, const char *ssid, int ssid_len)
 	struct ndis_essid req;
 
 	TRACEENTER2("");
-	memset(&req, 0, sizeof(req));
+	if (ssid_len > NDIS_ESSID_MAX_SIZE)
+		return -EINVAL;
 
+	memset(&req, 0, sizeof(req));
 	if (ssid_len == 0)
 		req.length = 1;
 	else {
-		if (ssid_len > NDIS_ESSID_MAX_SIZE)
-			return -EINVAL;
-
 		req.length = ssid_len;
 		memcpy(&req.essid, ssid, req.length);
 		DBGTRACE2("ssid = '%s'", req.essid);
@@ -121,26 +120,29 @@ static int set_assoc_params(struct wrap_ndis_device *wnd)
 	return 0;
 }
 
-
 static int iw_set_essid(struct net_device *dev, struct iw_request_info *info,
 			union iwreq_data *wrqu, char *extra)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
 	char ssid[IW_ESSID_MAX_SIZE];
+	int length;
 
 	TRACEENTER2("");
-	memset(ssid, 0, sizeof(ssid));
-	/* iwconfig adds 1 to the actual length */
-	/* there is no way to turn off essid other than to set to
-	 * random bytes; instead, we use off to mean any */
-	if (wrqu->essid.flags)
-		wrqu->essid.length--;
-	else
-		wrqu->essid.length = 0;
-
 	if (wrqu->essid.length > IW_ESSID_MAX_SIZE)
 		TRACEEXIT2(return -EINVAL);
 
+	/* there is no way to turn off essid other than to set to
+	 * random bytes; instead, we use off to mean any */
+	if (wrqu->essid.flags) {
+		/* wireless-tools prior to version 20 add extra 1, and
+		 * later than 20 don't! Deal with that mess */
+		for (length = 0; length < wrqu->essid.length; length++)
+			if (!extra[length])
+				break;
+	} else
+		length = 0;
+
+	memset(ssid, 0, sizeof(ssid));
 	if (wnd->iw_auth_set) {
 		int ret = set_assoc_params(wnd);
 		wnd->iw_auth_set = 0;
@@ -148,8 +150,8 @@ static int iw_set_essid(struct net_device *dev, struct iw_request_info *info,
 			TRACEEXIT2(return ret);
 	}
 
-	memcpy(ssid, extra, wrqu->essid.length);
-	if (set_essid(wnd, ssid, wrqu->essid.length))
+	memcpy(ssid, extra, length);
+	if (set_essid(wnd, ssid, length))
 		TRACEEXIT2(return -EINVAL);
 
 	TRACEEXIT2(return 0);
@@ -186,12 +188,19 @@ int set_infra_mode(struct wrap_ndis_device *wnd,
 	unsigned int i;
 
 	TRACEENTER2("");
+	res = miniport_query_int(wnd, OID_802_11_INFRASTRUCTURE_MODE,
+				 &wnd->infrastructure_mode);
+	if (res != NDIS_STATUS_SUCCESS) {
+		WARNING("getting operating mode to failed (%08X)", res);
+		TRACEEXIT2(return -EINVAL);
+	}
 	if (wnd->infrastructure_mode == mode)
 		TRACEEXIT2(return 0);
 
 	res = miniport_set_int(wnd, OID_802_11_INFRASTRUCTURE_MODE, mode);
 	if (res) {
-		WARNING("setting operating mode failed (%08X)", res);
+		WARNING("setting operating mode to %d failed (%08X)",
+			mode, res);
 		TRACEEXIT2(return -EINVAL);
 	}
 
@@ -343,33 +352,34 @@ static int iw_set_freq(struct net_device *dev, struct iw_request_info *info,
 	struct ndis_configuration req;
 
 	TRACEENTER2("");
+	/* this OID is valid only when not associated */
+	if (netif_carrier_ok(wnd->net_dev))
+		TRACEEXIT2(return 0);
 	memset(&req, 0, sizeof(req));
 	res = miniport_query_info(wnd, OID_802_11_CONFIGURATION,
 				  &req, sizeof(req));
 	if (res) {
 		WARNING("getting configuration failed (%08X)", res);
-		TRACEEXIT2(return -EOPNOTSUPP);
+		TRACEEXIT2(return 0);
 	}
 
 	if (wrqu->freq.m < 1000 && wrqu->freq.e == 0) {
 		if (wrqu->freq.m >= 1 &&
-		    wrqu->freq.m <= (sizeof(freq_chan)/sizeof(freq_chan[0])))
+		    wrqu->freq.m <= (sizeof(freq_chan) / sizeof(freq_chan[0])))
 			req.ds_config = freq_chan[wrqu->freq.m - 1] * 1000;
 		else
 			return -EINVAL;
 	} else {
 		int i;
-		for (req.ds_config = wrqu->freq.m, i = wrqu->freq.e ;
-		     i > 0 ; i--)
+		req.ds_config = wrqu->freq.m;
+		for (i = wrqu->freq.e; i > 0; i--)
 			req.ds_config *= 10;
 		req.ds_config /= 1000;
 	}
-	res = miniport_set_info(wnd, OID_802_11_CONFIGURATION, &req,
-				sizeof(req));
-	if (res) {
+	res = miniport_set_info(wnd, OID_802_11_CONFIGURATION,
+				&req, sizeof(req));
+	if (res)
 		WARNING("setting configuration failed (%08X)", res);
-		return -EINVAL;
-	}
 	return 0;
 }
 
@@ -592,13 +602,12 @@ int get_ap_address(struct wrap_ndis_device *wnd, mac_address ap_addr)
 {
 	NDIS_STATUS res;
 
-	TRACEENTER2("%d", wnd->link_status);
 	res = NDIS_STATUS_ADAPTER_NOT_READY;
 	/* this OID is valid only when associated */
-	if (wnd->link_status)
+	if (netif_carrier_ok(wnd->net_dev))
 		res = miniport_query_info(wnd, OID_802_11_BSSID, ap_addr,
 					  ETH_ALEN);
-	DBGTRACE2(MACSTR, MAC2STR(ap_addr));
+	DBGTRACE2(MACSTRSEP, MAC2STR(ap_addr));
 	if (res) {
 		DBGTRACE2("res: %08X", res);
 		memset(ap_addr, 0x0, ETH_ALEN);
@@ -631,7 +640,7 @@ static int iw_set_ap_address(struct net_device *dev,
 
 	TRACEENTER2("");
 	memcpy(ap_addr, wrqu->ap_addr.sa_data, ETH_ALEN);
-	DBGTRACE2(MACSTR, MAC2STR(ap_addr));
+	DBGTRACE2(MACSTRSEP, MAC2STR(ap_addr));
 	res = miniport_set_info(wnd, OID_802_11_BSSID, ap_addr, ETH_ALEN);
 	/* user apps may set ap's mac address, which is not required;
 	 * they may fail to work if this function fails, so return
@@ -802,7 +811,7 @@ int add_wep_key(struct wrap_ndis_device *wnd, char *key, int key_len,
 		if (res)
 			WARNING("encryption couldn't be enabled (%08X)", res);
 	}
-	DBGTRACE2("key %d: " MACSTR, index, MAC2STR(key));
+	DBGTRACE2("key %d: " MACSTRSEP, index, MAC2STR(key));
 	res = miniport_set_info(wnd, OID_802_11_ADD_WEP, &ndis_key,
 				sizeof(ndis_key));
 	if (res) {
@@ -1561,9 +1570,12 @@ static int iw_set_auth(struct net_device *dev,
 	case IW_AUTH_DROP_UNENCRYPTED:
 	case IW_AUTH_RX_UNENCRYPTED_EAPOL:
 	case IW_AUTH_PRIVACY_INVOKED:
+		DBGTRACE2("%d not implemented",
+			  wrqu->param.flags & IW_AUTH_INDEX);
 		/* TODO */
 		break;
 	default:
+		WARNING("invalid cmd %d", wrqu->param.flags & IW_AUTH_INDEX);
 		return -EOPNOTSUPP;
 	}
 	return 0;
@@ -1593,6 +1605,7 @@ static int iw_get_auth(struct net_device *dev,
 		wrqu->param.value = wnd->iw_auth_80211_auth_alg;
 		break;
 	default:
+		WARNING("invalid cmd %d", wrqu->param.flags & IW_AUTH_INDEX);
 		return -EOPNOTSUPP;
 	}
 	return 0;
@@ -1621,11 +1634,11 @@ static int iw_set_encodeext(struct net_device *dev,
 		return -EINVAL;
 
 	if (copy_from_user(&ext, wrqu->encoding.pointer, sizeof(ext)))
-		TRACEEXIT2(return -1);
+		TRACEEXIT2(return -EFAULT);
 
 	if (copy_from_user(key, wrqu->encoding.pointer + sizeof(ext),
 			   ext.key_len))
-		TRACEEXIT2(return -1);
+		TRACEEXIT2(return -EFAULT);
 
 	if (ext.alg == WPA_ALG_WEP) {
 		if (!test_bit(Ndis802_11Encryption1Enabled, &wnd->capa.encr))
@@ -1660,7 +1673,7 @@ static int iw_set_encodeext(struct net_device *dev,
 	}
 
 	addr = ext.addr.sa_data;
-	DBGTRACE2("infra_mode = %d, addr = " MACSTR,
+	DBGTRACE2("infra_mode = %d, addr = " MACSTRSEP,
 		  wnd->infrastructure_mode, MAC2STR(addr));
 
 	if (memcmp(addr, "\xff\xff\xff\xff\xff\xff", ETH_ALEN) == 0) {
@@ -1675,7 +1688,7 @@ static int iw_set_encodeext(struct net_device *dev,
 		memcpy(&ndis_key.bssid, addr, ETH_ALEN);
 	}
 
-	DBGTRACE2("bssid " MACSTR, MAC2STR(ndis_key.bssid));
+	DBGTRACE2("bssid " MACSTRSEP, MAC2STR(ndis_key.bssid));
 
 	if (ext.ext_flags & IW_ENCODE_EXT_SET_TX_KEY)
 		ndis_key.index |= (1 << 31);
@@ -2017,15 +2030,15 @@ static int wpa_set_key(struct net_device *dev, struct iw_request_info *info,
 	else
 		size = sizeof(wpa_key);
 	if (copy_from_user(&wpa_key, wrqu->data.pointer, size))
-		TRACEEXIT2(return -1);
+		TRACEEXIT2(return -EFAULT);
 	if (wpa_key.addr && copy_from_user(&addr, wpa_key.addr, ETH_ALEN))
-		TRACEEXIT2(return -1);
+		TRACEEXIT2(return -EFAULT);
 
 	if (wpa_key.seq && copy_from_user(&seq, wpa_key.seq, wpa_key.seq_len))
-		TRACEEXIT2(return -1);
+		TRACEEXIT2(return -EFAULT);
 
 	if (wpa_key.key && copy_from_user(&key, wpa_key.key, wpa_key.key_len))
-		TRACEEXIT2(return -1);
+		TRACEEXIT2(return -EFAULT);
 
 	TRACEENTER2("alg = %d, key_index = %d",
 		    wpa_key.alg, wpa_key.key_index);
@@ -2062,7 +2075,7 @@ static int wpa_set_key(struct net_device *dev, struct iw_request_info *info,
 
 		ndis_key.index |= 1 << 29;
 	}
-	DBGTRACE2("infra_mode = %d, key.addr = %p, addr = " MACSTR,
+	DBGTRACE2("infra_mode = %d, key.addr = %p, addr = " MACSTRSEP,
 		  wnd->infrastructure_mode, wpa_key.addr, MAC2STR(addr));
 	if (wpa_key.addr == NULL ||
 	    memcmp(addr, "\xff\xff\xff\xff\xff\xff", ETH_ALEN) == 0) {
@@ -2076,7 +2089,7 @@ static int wpa_set_key(struct net_device *dev, struct iw_request_info *info,
 		ndis_key.index |= (1 << 30);
 		memcpy(&ndis_key.bssid, addr, ETH_ALEN);
 	}
-	DBGTRACE2("bssid " MACSTR, MAC2STR(ndis_key.bssid));
+	DBGTRACE2("bssid " MACSTRSEP, MAC2STR(ndis_key.bssid));
 
 	if (wpa_key.set_tx)
 		ndis_key.index |= (1 << 31);
@@ -2143,10 +2156,10 @@ static int wpa_associate(struct net_device *dev, struct iw_request_info *info,
 		size = min((size_t)wrqu->data.length, sizeof(wpa_assoc_info));
 
 	if (copy_from_user(&wpa_assoc_info, wrqu->data.pointer, size))
-		TRACEEXIT2(return -1);
+		TRACEEXIT2(return -EFAULT);
 	if (copy_from_user(&ssid, wpa_assoc_info.ssid,
 			   wpa_assoc_info.ssid_len))
-		TRACEEXIT2(return -1);
+		TRACEEXIT2(return -EFAULT);
 
 	if (wpa_assoc_info.mode == IEEE80211_MODE_IBSS)
 		infra_mode = Ndis802_11IBSS;
