@@ -22,14 +22,14 @@
 #define MAX_ALLOCATED_NDIS_PACKETS 20
 #define MAX_ALLOCATED_NDIS_BUFFERS 20
 
-extern NT_SPIN_LOCK loader_lock;
-
 static struct workqueue_struct *ndis_wq;
 static void ndis_worker(void *dummy);
 static struct work_struct ndis_work;
 static struct nt_list ndis_worker_list;
 static NT_SPIN_LOCK ndis_work_list_lock;
 
+extern struct semaphore loader_mutex;
+
 /* ndis_init is called once when module is loaded */
 int ndis_init(void)
 {
@@ -62,7 +62,8 @@ void ndis_exit_device(struct wrap_ndis_device *wnd)
 	/* TI driver doesn't call NdisMDeregisterInterrupt during halt! */
 	if (wnd->ndis_irq)
 		NdisMDeregisterInterrupt(wnd->ndis_irq);
-	nt_spin_lock(&loader_lock);
+	if (down_interruptible(&loader_mutex))
+		WARNING("couldn't obtain loader_mutex");
 	nt_list_for_each_entry(setting, &wnd->wd->settings, list) {
 		struct ndis_configuration_parameter *param;
 		param = setting->encoded;
@@ -73,7 +74,7 @@ void ndis_exit_device(struct wrap_ndis_device *wnd)
 			setting->encoded = NULL;
 		}
 	}
-	nt_spin_unlock(&loader_lock);
+	up(&loader_mutex);
 }
 
 wstdcall void WIN_FUNC(NdisInitializeWrapper,4)
@@ -86,14 +87,14 @@ wstdcall void WIN_FUNC(NdisInitializeWrapper,4)
 }
 
 wstdcall void WIN_FUNC(NdisTerminateWrapper,2)
-	(struct device_object *dev_obj, void *SystemSpecific1)
+	(struct device_object *dev_obj, void *system_specific)
 {
 	TRACEEXIT1(return);
 }
 
 wstdcall NDIS_STATUS WIN_FUNC(NdisMRegisterMiniport,3)
 	(struct driver_object *drv_obj,
-	 struct miniport_char *miniport_char, UINT char_len)
+	 struct miniport_char *miniport_char, UINT length)
 {
 	int min_length;
 	struct wrap_driver *wrap_driver;
@@ -102,7 +103,7 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisMRegisterMiniport,3)
 	min_length = ((char *)&miniport_char->co_create_vc) -
 		((char *)miniport_char);
 
-	TRACEENTER1("%p %p %d", drv_obj, miniport_char, char_len);
+	TRACEENTER1("%p %p %d", drv_obj, miniport_char, length);
 
 	if (miniport_char->major_version < 4) {
 		ERROR("Driver is using ndis version %d which is too old.",
@@ -110,13 +111,13 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisMRegisterMiniport,3)
 		TRACEEXIT1(return NDIS_STATUS_BAD_VERSION);
 	}
 
-	if (char_len < min_length) {
-		ERROR("Characteristics length %d is too small", char_len);
+	if (length < min_length) {
+		ERROR("Characteristics length %d is too small", length);
 		TRACEEXIT1(return NDIS_STATUS_BAD_CHARACTERISTICS);
 	}
 
 	DBGTRACE1("%d.%d, %d, %u", miniport_char->major_version,
-		  miniport_char->minor_version, char_len,
+		  miniport_char->minor_version, length,
 		  (u32)sizeof(struct miniport_char));
 	wrap_driver =
 		IoGetDriverObjectExtension(drv_obj,
@@ -133,8 +134,8 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisMRegisterMiniport,3)
 	wrap_driver->ndis_driver = ndis_driver;
 	TRACEENTER1("driver: %p", ndis_driver);
 	memcpy(&ndis_driver->miniport, miniport_char,
-	       char_len > sizeof(*miniport_char) ?
-	       sizeof(*miniport_char) : char_len);
+	       length > sizeof(*miniport_char) ?
+	       sizeof(*miniport_char) : length);
 
 	DBG_BLOCK(2) {
 		int i;
@@ -224,9 +225,6 @@ wstdcall void WIN_FUNC(NdisFreeMemory,3)
 	ExFreePool(addr);
 }
 
-/*
- * This function should not be wstdcall because it's a variable args function.
- */
 noregparm void WIN_FUNC(NdisWriteErrorLogEntry,12)
 	(struct driver_object *drv_obj, ULONG error, ULONG count, ...)
 {
@@ -500,11 +498,12 @@ wstdcall void WIN_FUNC(NdisReadConfiguration,5)
 	DBGTRACE3("%d, %s", type, ansi.buf);
 	keyname = ansi.buf;
 
-	nt_spin_lock(&loader_lock);
+	if (down_interruptible(&loader_mutex))
+		WARNING("couldn't obtain loader_mutex");
 	nt_list_for_each_entry(setting, &nmb->wnd->wd->settings, list) {
 		if (strnicmp(keyname, setting->name, ansi.length) == 0) {
 			DBGTRACE2("setting %s='%s'", keyname, setting->value);
-			nt_spin_unlock(&loader_lock);
+			up(&loader_mutex);
 			*param = ndis_encode_setting(setting, type);
 			if (*param)
 				*status = NDIS_STATUS_SUCCESS;
@@ -515,7 +514,7 @@ wstdcall void WIN_FUNC(NdisReadConfiguration,5)
 			TRACEEXIT2(return);
 		}
 	}
-	nt_spin_unlock(&loader_lock);
+	up(&loader_mutex);
 	DBGTRACE2("setting %s not found (type:%d)", keyname, type);
 	*status = NDIS_STATUS_FAILURE;
 	RtlFreeAnsiString(&ansi);
@@ -538,10 +537,11 @@ wstdcall void WIN_FUNC(NdisWriteConfiguration,4)
 	keyname = ansi.buf;
 	DBGTRACE2("%s", keyname);
 
-	nt_spin_lock(&loader_lock);
+	if (down_interruptible(&loader_mutex))
+		WARNING("couldn't obtain loader_mutex");
 	nt_list_for_each_entry(setting, &nmb->wnd->wd->settings, list) {
 		if (strnicmp(keyname, setting->name, ansi.length) == 0) {
-			nt_spin_unlock(&loader_lock);
+			up(&loader_mutex);
 			if (ndis_decode_setting(setting, param))
 				*status = NDIS_STATUS_FAILURE;
 			else
@@ -550,7 +550,7 @@ wstdcall void WIN_FUNC(NdisWriteConfiguration,4)
 			TRACEEXIT2(return);
 		}
 	}
-	nt_spin_unlock(&loader_lock);
+	up(&loader_mutex);
 	setting = kmalloc(sizeof(*setting), GFP_KERNEL);
 	if (setting) {
 		memset(setting, 0, sizeof(*setting));
@@ -562,10 +562,10 @@ wstdcall void WIN_FUNC(NdisWriteConfiguration,4)
 			*status = NDIS_STATUS_FAILURE;
 		else {
 			*status = NDIS_STATUS_SUCCESS;
-			nt_spin_lock(&loader_lock);
-			InsertTailList(&nmb->wnd->wd->settings,
-				       &setting->list);
-			nt_spin_unlock(&loader_lock);
+			if (down_interruptible(&loader_mutex))
+				WARNING("couldn't obtain loader_mutex");
+			InsertTailList(&nmb->wnd->wd->settings, &setting->list);
+			up(&loader_mutex);
 		}
 	} else
 		*status = NDIS_STATUS_RESOURCES;
@@ -690,25 +690,25 @@ wstdcall ULONG WIN_FUNC(NdisWritePciSlotInformation,5)
 wstdcall void WIN_FUNC(NdisReadPortUchar,3)
 	(struct ndis_miniport_block *nmb, ULONG port, char *data)
 {
-	*data = READ_PORT_UCHAR(port);
+	*data = inb(port);
 }
 
 wstdcall void WIN_FUNC(NdisImmediateReadPortUchar,3)
 	(struct ndis_miniport_block *nmb, ULONG port, char *data)
 {
-	*data = READ_PORT_UCHAR(port);
+	*data = inb(port);
 }
 
 wstdcall void WIN_FUNC(NdisWritePortUchar,3)
 	(struct ndis_miniport_block *nmb, ULONG port, char data)
 {
-	WRITE_PORT_UCHAR(port, data);
+	outb(data, port);
 }
 
 wstdcall void WIN_FUNC(NdisImmediateWritePortUchar,3)
 	(struct ndis_miniport_block *nmb, ULONG port, char data)
 {
-	WRITE_PORT_UCHAR(port, data);
+	outb(data, port);
 }
 
 wstdcall void WIN_FUNC(NdisMQueryAdapterResources,4)
@@ -738,8 +738,8 @@ wstdcall void WIN_FUNC(NdisMQueryAdapterResources,4)
 			count = list->count;
 		} else {
 			UINT n = sizeof(*list);
-			count = 0;
-			while (++count < list->count && n < *size)
+			count = 1;
+			while (count++ < list->count && n < *size)
 				n += sizeof(list->partial_descriptors);
 			*size = n;
 		}
@@ -773,7 +773,6 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisMMapIoSpace,4)
 		ERROR("ioremap failed");
 		TRACEEXIT2(return NDIS_STATUS_FAILURE);
 	}
-
 	wnd->mem_start = phy_addr;
 	wnd->mem_end = phy_addr + len;
 	DBGTRACE2("%p", *virt);
@@ -852,30 +851,27 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisMAllocateMapRegisters,5)
 	struct wrap_ndis_device *wnd = nmb->wnd;
 	TRACEENTER2("%d %d %d %d", dmachan, dmasize, basemap, max_buf_size);
 
-	if (max_buf_size > PAGE_SIZE)
-	    WARNING("buffer size too big: %d", max_buf_size);
-
-//	if (basemap > 64)
-//		return NDIS_STATUS_RESOURCES;
+	if (basemap > 64) {
+		WARNING("invalid request: %d", basemap);
+		TRACEEXIT1(return NDIS_STATUS_RESOURCES);
+	}
 
 	/* TODO: should we allocate max_buf_size % PAGE_SIZE + 1
 	 * entries for each map? */
 	if (max_buf_size > PAGE_SIZE)
 		WARNING("buffer size too big: %u", max_buf_size);
 
-	if (wnd->dma_map_count > 0) {
+	if (wnd->tx_dma_count > 0) {
 		DBGTRACE2("%s: map registers already allocated: %u",
-			  wnd->net_dev->name, wnd->dma_map_count);
+			  wnd->net_dev->name, wnd->tx_dma_count);
 		TRACEEXIT2(return NDIS_STATUS_RESOURCES);
 	}
-
-	wnd->dma_map_count = basemap;
-	wnd->dma_map_addr = kmalloc(basemap * sizeof(dma_addr_t),
+	wnd->tx_dma_count = basemap;
+	wnd->tx_dma_addr = kmalloc(basemap * sizeof(dma_addr_t),
 				   GFP_KERNEL);
-	if (!wnd->dma_map_addr)
+	if (!wnd->tx_dma_addr)
 		TRACEEXIT2(return NDIS_STATUS_RESOURCES);
-	memset(wnd->dma_map_addr, 0, basemap * sizeof(dma_addr_t));
-
+	memset(wnd->tx_dma_addr, 0, basemap * sizeof(dma_addr_t));
 	TRACEEXIT2(return NDIS_STATUS_SUCCESS);
 }
 
@@ -885,12 +881,87 @@ wstdcall void WIN_FUNC(NdisMFreeMapRegisters,1)
 	struct wrap_ndis_device *wnd = nmb->wnd;
 	TRACEENTER2("wnd: %p", wnd);
 
-	if (wnd->dma_map_addr != NULL)
-		kfree(wnd->dma_map_addr);
-	wnd->dma_map_count = 0;
+	if (wnd->tx_dma_addr) {
+		u8 i;
+		for (i = 0; i < wnd->tx_dma_count; i++) {
+			if (wnd->tx_dma_addr[i])
+				WARNING("%s: dma addr %p not freed by "
+					"Windows driver", wnd->net_dev->name,
+					(void *)wnd->tx_dma_addr[i]);
+		}
+		kfree(wnd->tx_dma_addr);
+		wnd->tx_dma_addr = NULL;
+	}
+	wnd->tx_dma_count = 0;
 	TRACEEXIT2(return);
 }
 
+wstdcall void WIN_FUNC(NdisMStartBufferPhysicalMapping,6)
+	(struct ndis_miniport_block *nmb, ndis_buffer *buf,
+	 ULONG index, BOOLEAN write_to_dev,
+	 struct ndis_phy_addr_unit *phy_addr_array, UINT *array_size)
+{
+	struct wrap_ndis_device *wnd = nmb->wnd;
+	TRACEENTER3("index: %u", index);
+
+	if (wnd->use_sg_dma)
+		WARNING("buffer %p must have been mapped already", buf);
+	if (!write_to_dev) {
+		ERROR("invalid dma direction (%d)", write_to_dev);
+		*array_size = 0;
+		return;
+	}
+	if (index >= wnd->tx_dma_count) {
+		ERROR("invalid map register (%u >= %u)",
+		      index, wnd->tx_dma_count);
+		*array_size = 0;
+		return;
+	}
+	if (wnd->tx_dma_addr[index] != 0) {
+		*array_size = 1;
+		return;
+	}
+	DBGTRACE3("%p, %p, %u", buf, MmGetSystemAddressForMdl(buf),
+		  MmGetMdlByteCount(buf));
+	DBG_BLOCK(4) {
+		dump_bytes(__FUNCTION__, MmGetSystemAddressForMdl(buf),
+			   MmGetMdlByteCount(buf));
+	}
+	phy_addr_array[0].phy_addr =
+		PCI_DMA_MAP_SINGLE(wnd->wd->pci.pdev,
+				   MmGetSystemAddressForMdl(buf),
+				   MmGetMdlByteCount(buf), PCI_DMA_TODEVICE);
+	phy_addr_array[0].length = MmGetMdlByteCount(buf);
+	*array_size = 1;
+	DBGTRACE4("%Lx, %d, %d", phy_addr_array[0].phy_addr,
+		  phy_addr_array[0].length, index);
+	wnd->tx_dma_addr[index] = phy_addr_array[0].phy_addr;
+}
+
+wstdcall void WIN_FUNC(NdisMCompleteBufferPhysicalMapping,3)
+	(struct ndis_miniport_block *nmb, ndis_buffer *buf, ULONG index)
+{
+	struct wrap_ndis_device *wnd = nmb->wnd;
+	TRACEENTER3("%p, %p %u (%u)", wnd, buf, index, wnd->tx_dma_count);
+
+	if (wnd->use_sg_dma)
+		WARNING("buffer %p may have been unmapped already", buf);
+	if (index >= wnd->tx_dma_count) {
+		ERROR("invalid map register (%u >= %u)",
+		      index, wnd->tx_dma_count);
+		return;
+	}
+	DBGTRACE4("%lx, %d, %d", (unsigned long)wnd->tx_dma_addr[index],
+		  MmGetMdlByteCount(buf), index);
+	if (wnd->tx_dma_addr[index] == 0) {
+//		ERROR("map register not used (%lu)", index);
+		return;
+	}
+	PCI_DMA_UNMAP_SINGLE(wnd->wd->pci.pdev, wnd->tx_dma_addr[index],
+			     MmGetMdlByteCount(buf), PCI_DMA_TODEVICE);
+	wnd->tx_dma_addr[index] = 0;
+}
+
 wstdcall void WIN_FUNC(NdisMAllocateSharedMemory,5)
 	(struct ndis_miniport_block *nmb, ULONG size,
 	 BOOLEAN cached, void **virt, NDIS_PHY_ADDRESS *phys)
@@ -898,41 +969,23 @@ wstdcall void WIN_FUNC(NdisMAllocateSharedMemory,5)
 	dma_addr_t dma_addr;
 	struct wrap_device *wd = nmb->wnd->wd;
 
-	TRACEENTER3("count: %d, size: %u, cached: %d",
-		    nmb->wnd->dma_map_count, size, cached);
+	TRACEENTER3("size: %u, cached: %d", size, cached);
 	*virt = PCI_DMA_ALLOC_COHERENT(wd->pci.pdev, size, &dma_addr);
 	if (!*virt)
-		ERROR("couldn't allocate %d bytes of %scached DMA memory",
-		      size, cached ? "" : "un-");
+		WARNING("couldn't allocate %d bytes of %scached DMA memory",
+			size, cached ? "" : "un-");
 	*phys = dma_addr;
 	TRACEEXIT3(return);
 }
 
-static void ndis_worker(void *dummy)
+wstdcall void WIN_FUNC(NdisMFreeSharedMemory,5)
+	(struct ndis_miniport_block *nmb, ULONG size, BOOLEAN cached,
+	 void *virt, NDIS_PHY_ADDRESS addr)
 {
-	KIRQL irql;
-	struct ndis_work_entry *ndis_work_entry;
-	struct nt_list *ent;
-	struct ndis_work_item *ndis_work_item;
-
-	WORKENTER("");
-	while (1) {
-		irql = nt_spin_lock_irql(&ndis_work_list_lock, DISPATCH_LEVEL);
-		ent = RemoveHeadList(&ndis_worker_list);
-		nt_spin_unlock_irql(&ndis_work_list_lock, irql);
-		if (!ent)
-			break;
-		ndis_work_entry = container_of(ent, struct ndis_work_entry,
-					       list);
-		ndis_work_item = ndis_work_entry->ndis_work_item;
-		WORKTRACE("%p: %p, %p", ndis_work_item,
-			  ndis_work_item->func, ndis_work_item->ctx);
-		LIN2WIN2(ndis_work_item->func, ndis_work_item,
-			 ndis_work_item->ctx);
-		WORKTRACE("%p done", ndis_work_item);
-		kfree(ndis_work_entry);
-	}
-	WORKEXIT(return);
+	struct wrap_device *wd = nmb->wnd->wd;
+	TRACEENTER3("");
+	PCI_DMA_FREE_COHERENT(wd->pci.pdev, size, virt, addr);
+	TRACEEXIT3(return);
 }
 
 wstdcall void alloc_shared_memory_async(void *arg1, void *arg2)
@@ -980,16 +1033,6 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisMAllocateSharedMemoryAsync,4)
 	TRACEEXIT3(return NDIS_STATUS_PENDING);
 }
 
-wstdcall void WIN_FUNC(NdisMFreeSharedMemory,5)
-	(struct ndis_miniport_block *nmb, ULONG size, BOOLEAN cached,
-	 void *virt, NDIS_PHY_ADDRESS addr)
-{
-	struct wrap_device *wd = nmb->wnd->wd;
-	TRACEENTER3("");
-	PCI_DMA_FREE_COHERENT(wd->pci.pdev, size, virt, addr);
-	TRACEEXIT3(return);
-}
-
 /* Some drivers allocate NDIS_BUFFER (aka MDL) very often; instead of
  * allocating and freeing with kernel functions, we chain them into
  * ndis_buffer_pool. When an MDL is freed, it is added to the list of
@@ -1003,6 +1046,7 @@ wstdcall void WIN_FUNC(NdisMFreeSharedMemory,5)
  * pass the same pool_handle in other buffer functions, this should
  * work. Sadly, though, NdisFreeBuffer doesn't pass the pool_handle,
  * so we use 'process' field of MDL to store pool_handle. */
+
 wstdcall void WIN_FUNC(NdisAllocateBufferPool,3)
 	(NDIS_STATUS *status, struct ndis_buffer_pool **pool_handle,
 	 UINT num_descr)
@@ -1030,7 +1074,6 @@ wstdcall void WIN_FUNC(NdisAllocateBuffer,5)
 	 struct ndis_buffer_pool *pool, void *virt, UINT length)
 {
 	ndis_buffer *descr;
-	KIRQL irql;
 
 	TRACEENTER4("pool: %p, allocated: %d",
 		    pool, pool->num_allocated_descr);
@@ -1038,21 +1081,20 @@ wstdcall void WIN_FUNC(NdisAllocateBuffer,5)
 		*status = NDIS_STATUS_FAILURE;
 		TRACEEXIT4(return);
 	}
-	irql = nt_spin_lock_irql(&pool->lock, DISPATCH_LEVEL);
-	if (pool->num_allocated_descr > pool->max_descr)
-		DBGTRACE4("pool %p is full: %d(%d)", pool,
-			  pool->num_allocated_descr, pool->max_descr);
-	if (pool->free_descr) {
+	DBG_BLOCK(2) {
+		if (pool->num_allocated_descr > pool->max_descr)
+			WARNING("pool %p is full: %d(%d)", pool,
+				pool->num_allocated_descr, pool->max_descr);
+	}
+	descr = atomic_remove_list_head(pool->free_descr, oldhead->next);
+	if (descr) {
 		typeof(descr->flags) flags;
-		descr = pool->free_descr;
-		pool->free_descr = descr->next;
 		flags = descr->flags;
-		memset(descr, 0, sizeof(*descr));
 		MmInitializeMdl(descr, virt, length);
 		if (flags & MDL_CACHE_ALLOCATED)
 			descr->flags = MDL_CACHE_ALLOCATED;
+		descr->flags |= MDL_SOURCE_IS_NONPAGED_POOL;
 	} else {
-		nt_spin_unlock_irql(&pool->lock, irql);
 		descr = allocate_init_mdl(virt, length);
 		if (!descr) {
 			WARNING("couldn't allocate buffer");
@@ -1061,14 +1103,10 @@ wstdcall void WIN_FUNC(NdisAllocateBuffer,5)
 		}
 		DBGTRACE4("allocated buffer %p for %p, %d",
 			  descr, virt, length);
-		irql = nt_spin_lock_irql(&pool->lock, DISPATCH_LEVEL);
-		pool->num_allocated_descr++;
+		atomic_inc_var(pool->num_allocated_descr);
+		MmBuildMdlForNonPagedPool(descr);
 	}
-	MmBuildMdlForNonPagedPool(descr);
-	/* NdisFreeBuffer doesn't pass pool, so we store pool
-	 * in unused field 'process' */
-	descr->process = pool;
-	nt_spin_unlock_irql(&pool->lock, irql);
+	descr->pool = pool;
 	*buffer = descr;
 	*status = NDIS_STATUS_SUCCESS;
 	DBGTRACE4("buffer: %p", descr);
@@ -1079,32 +1117,22 @@ wstdcall void WIN_FUNC(NdisFreeBuffer,1)
 	(ndis_buffer *descr)
 {
 	struct ndis_buffer_pool *pool;
-	KIRQL irql;
 
 	TRACEENTER4("descr: %p", descr);
-	if (!descr) {
-		ERROR("buffer is NULL");
+	if (!descr || !descr->pool) {
+		ERROR("invalid buffer");
 		TRACEEXIT4(return);
 	}
-	pool = descr->process;
-	if (!pool) {
-		ERROR("pool for descriptor %p is invalid", descr);
-		TRACEEXIT4(return);
-	}
-	irql = nt_spin_lock_irql(&pool->lock, DISPATCH_LEVEL);
+	pool = descr->pool;
 	if (pool->num_allocated_descr > MAX_ALLOCATED_NDIS_BUFFERS) {
-		/* NB NB NB: set mdl's 'process' field to NULL before
+		/* NB NB NB: set mdl's 'pool' field to NULL before
 		 * calling free_mdl; otherwise free_mdl calls
 		 * NdisFreeBuffer causing deadlock (for spinlock) */
-		pool->num_allocated_descr--;
-		descr->process = NULL;
-		nt_spin_unlock_irql(&pool->lock, irql);
+		atomic_dec_var(pool->num_allocated_descr);
+		descr->pool = NULL;
 		free_mdl(descr);
-	} else {
-		descr->next = pool->free_descr;
-		pool->free_descr = descr;
-		nt_spin_unlock_irql(&pool->lock, irql);
-	}
+	} else
+		atomic_insert_list_head(pool->free_descr, descr->next, descr);
 	TRACEEXIT4(return);
 }
 
@@ -1123,7 +1151,7 @@ wstdcall void WIN_FUNC(NdisFreeBufferPool,1)
 	cur = pool->free_descr;
 	while (cur) {
 		next = cur->next;
-		cur->process = NULL;
+		cur->pool = NULL;
 		free_mdl(cur);
 		cur = next;
 	}
@@ -1271,79 +1299,6 @@ wstdcall void WIN_FUNC(NdisGetFirstBufferFromPacket,6)
 					 NormalPagePriority);
 }
 
-wstdcall void WIN_FUNC(NdisMStartBufferPhysicalMapping,6)
-	(struct ndis_miniport_block *nmb, ndis_buffer *buf,
-	 ULONG map_index, BOOLEAN write_to_dev,
-	 struct ndis_phy_addr_unit *phy_addr_array, UINT *array_size)
-{
-	struct wrap_ndis_device *wnd = nmb->wnd;
-	TRACEENTER3("map_index: %u", map_index);
-
-	if (!write_to_dev) {
-		ERROR("invalid dma direction (%d)", write_to_dev);
-		*array_size = 0;
-		return;
-	}
-
-	if (map_index >= wnd->dma_map_count) {
-		ERROR("invalid map register (%u >= %u)",
-		      map_index, wnd->dma_map_count);
-		*array_size = 0;
-		return;
-	}
-
-	if (wnd->dma_map_addr[map_index] != 0) {
-		*array_size = 1;
-		return;
-	}
-
-	DBGTRACE3("%p, %p, %u", buf, MmGetSystemAddressForMdl(buf),
-		  MmGetMdlByteCount(buf));
-	DBG_BLOCK(4) {
-		dump_bytes(__FUNCTION__, MmGetSystemAddressForMdl(buf),
-			   MmGetMdlByteCount(buf));
-	}
-	// map buffer
-	phy_addr_array[0].phy_addr =
-		PCI_DMA_MAP_SINGLE(wnd->wd->pci.pdev,
-				   MmGetSystemAddressForMdl(buf),
-				   MmGetMdlByteCount(buf), PCI_DMA_TODEVICE);
-	phy_addr_array[0].length = MmGetMdlByteCount(buf);
-
-	*array_size = 1;
-	DBGTRACE4("%Lx, %d, %d", phy_addr_array[0].phy_addr,
-		  phy_addr_array[0].length, map_index);
-	// save map address
-	wnd->dma_map_addr[map_index] = phy_addr_array[0].phy_addr;
-}
-
-wstdcall void WIN_FUNC(NdisMCompleteBufferPhysicalMapping,3)
-	(struct ndis_miniport_block *nmb, ndis_buffer *buf, ULONG map_index)
-{
-	struct wrap_ndis_device *wnd = nmb->wnd;
-	TRACEENTER3("%p, %p %u (%u)", wnd, buf, map_index, wnd->dma_map_count);
-
-	if (map_index >= wnd->dma_map_count) {
-		ERROR("invalid map register (%u >= %u)",
-		      map_index, wnd->dma_map_count);
-		return;
-	}
-	DBGTRACE4("%lx, %d, %d", (unsigned long)wnd->dma_map_addr[map_index],
-		  MmGetMdlByteCount(buf), map_index);
-	if (wnd->dma_map_addr[map_index] == 0) {
-//		ERROR("map register not used (%lu)", map_index);
-		return;
-	}
-
-	// unmap buffer
-	PCI_DMA_UNMAP_SINGLE(wnd->wd->pci.pdev,
-			     wnd->dma_map_addr[map_index],
-			     MmGetMdlByteCount(buf), PCI_DMA_TODEVICE);
-
-	// clear map address
-	wnd->dma_map_addr[map_index] = 0;
-}
-
 wstdcall void WIN_FUNC(NdisAllocatePacketPoolEx,5)
 	(NDIS_STATUS *status, struct ndis_packet_pool **pool_handle,
 	 UINT num_descr, UINT overflowsize, UINT proto_rsvd_length)
@@ -1356,6 +1311,7 @@ wstdcall void WIN_FUNC(NdisAllocatePacketPoolEx,5)
 		*status = NDIS_STATUS_RESOURCES;
 		TRACEEXIT3(return);
 	}
+	memset(pool, 0, sizeof(*pool));
 	nt_spin_lock_init(&pool->lock);
 	pool->max_descr = num_descr;
 	pool->num_allocated_descr = 0;
@@ -1372,7 +1328,6 @@ wstdcall void WIN_FUNC(NdisAllocatePacketPool,4)
 	(NDIS_STATUS *status, struct ndis_packet_pool **pool_handle,
 	 UINT num_descr, UINT proto_rsvd_length)
 {
-	TRACEENTER3("");
 	NdisAllocatePacketPoolEx(status, pool_handle, num_descr, 0,
 				 proto_rsvd_length);
 	TRACEEXIT3(return);
@@ -1400,7 +1355,6 @@ wstdcall void WIN_FUNC(NdisFreePacketPool,1)
 	pool->num_used_descr = 0;
 	pool->free_descr = NULL;
 	nt_spin_unlock_irql(&pool->lock, irql);
-	DBGTRACE3("pool: %p", pool);
 	kfree(pool);
 	TRACEEXIT3(return);
 }
@@ -1408,14 +1362,7 @@ wstdcall void WIN_FUNC(NdisFreePacketPool,1)
 wstdcall UINT WIN_FUNC(NdisPacketPoolUsage,1)
 	(struct ndis_packet_pool *pool)
 {
-	UINT i;
-	KIRQL irql;
-
-	TRACEENTER4("");
-	irql = nt_spin_lock_irql(&pool->lock, DISPATCH_LEVEL);
-	i = pool->num_used_descr;
-	nt_spin_unlock_irql(&pool->lock, irql);
-	TRACEEXIT4(return i);
+	TRACEEXIT4(return pool->num_used_descr);
 }
 
 wstdcall void WIN_FUNC(NdisAllocatePacket,3)
@@ -1424,50 +1371,42 @@ wstdcall void WIN_FUNC(NdisAllocatePacket,3)
 {
 	struct ndis_packet *ndis_packet;
 	int packet_length;
-	KIRQL irql;
 
 	TRACEENTER4("pool: %p", pool);
 	if (!pool) {
 		*status = NDIS_STATUS_RESOURCES;
 		TRACEEXIT4(return);
 	}
+	DBG_BLOCK(2) {
+		if (pool->num_used_descr >= pool->max_descr)
+			WARNING("pool %p is full: %d(%d)", pool,
+				pool->num_used_descr, pool->max_descr);
+	}
 	/* packet has space for 1 byte in protocol_reserved field */
 	packet_length = sizeof(*ndis_packet) - 1 + pool->proto_rsvd_length +
 		sizeof(struct ndis_packet_oob_data);
-	irql = nt_spin_lock_irql(&pool->lock, DISPATCH_LEVEL);
-	if (pool->num_used_descr >= pool->max_descr)
-		DBGTRACE4("pool %p is full: %d(%d)", pool,
-			  pool->num_used_descr, pool->max_descr);
-	ndis_packet = NULL;
-	if (pool->free_descr) {
-		struct ndis_packet_oob_data *oob_data;
-		ndis_packet = pool->free_descr;
-		oob_data = NDIS_PACKET_OOB_DATA(ndis_packet);
-		pool->free_descr = oob_data->next;
-	} else {
-		nt_spin_unlock_irql(&pool->lock, irql);
+	ndis_packet =
+		atomic_remove_list_head(pool->free_descr,
+					(NDIS_PACKET_OOB_DATA(oldhead))->next);
+	if (!ndis_packet) {
 		ndis_packet = kmalloc(packet_length, gfp_irql());
 		if (!ndis_packet) {
 			WARNING("couldn't allocate packet");
 			*status = NDIS_STATUS_RESOURCES;
 			return;
 		}
-		DBGTRACE3("allocated packet: %p", ndis_packet);
-		irql = nt_spin_lock_irql(&pool->lock, DISPATCH_LEVEL);
-		pool->num_allocated_descr++;
+		atomic_inc_var(pool->num_allocated_descr);
 	}
-	pool->num_used_descr++;
+	DBGTRACE3("packet: %p", ndis_packet);
+	atomic_inc_var(pool->num_used_descr);
 	memset(ndis_packet, 0, packet_length);
 	ndis_packet->private.oob_offset = packet_length -
 		sizeof(struct ndis_packet_oob_data);
-	ndis_packet->private.packet_flags =
-		fPACKET_ALLOCATED_BY_NDIS;
-//		fPACKET_ALLOCATED_BY_NDIS | NDIS_PROTOCOL_ID_TCP_IP;
+	ndis_packet->private.packet_flags = fPACKET_ALLOCATED_BY_NDIS;
+//		| NDIS_PROTOCOL_ID_TCP_IP;
 	ndis_packet->private.pool = pool;
-	nt_spin_unlock_irql(&pool->lock, irql);
-
-	*status = NDIS_STATUS_SUCCESS;
 	*packet = ndis_packet;
+	*status = NDIS_STATUS_SUCCESS;
 	DBGTRACE4("packet: %p, pool: %p", ndis_packet, pool);
 	TRACEEXIT4(return);
 }
@@ -1483,7 +1422,6 @@ wstdcall void WIN_FUNC(NdisFreePacket,1)
 	(struct ndis_packet *descr)
 {
 	struct ndis_packet_pool *pool;
-	KIRQL irql;
 
 	TRACEENTER3("packet: %p, pool: %p", descr, descr->private.pool);
 	pool = descr->private.pool;
@@ -1491,20 +1429,14 @@ wstdcall void WIN_FUNC(NdisFreePacket,1)
 		ERROR("pool for descriptor %p is invalid", descr);
 		TRACEEXIT4(return);
 	}
-	irql = nt_spin_lock_irql(&pool->lock, DISPATCH_LEVEL);
-	pool->num_used_descr--;
+	atomic_dec_var(pool->num_used_descr);
 	if (pool->num_allocated_descr > MAX_ALLOCATED_NDIS_PACKETS) {
-		pool->num_allocated_descr--;
-		nt_spin_unlock_irql(&pool->lock, irql);
 		kfree(descr);
+		atomic_dec_var(pool->num_allocated_descr);
 	} else {
 		struct ndis_packet_oob_data *oob_data;
 		oob_data = NDIS_PACKET_OOB_DATA(descr);
-		descr->private.buffer_head = NULL;
-		descr->private.valid_counts = FALSE;
-		oob_data->next = pool->free_descr;
-		pool->free_descr = descr;
-		nt_spin_unlock_irql(&pool->lock, irql);
+		atomic_insert_list_head(pool->free_descr, oob_data->next, descr);
 	}
 	TRACEEXIT4(return);
 }
@@ -1747,7 +1679,7 @@ wstdcall void WIN_FUNC(NdisReadNetworkAddress,4)
 	int ret;
 
 	TRACEENTER1("");
-	RtlInitAnsiString(&ansi, "mac_address");
+	RtlInitAnsiString(&ansi, "NetworkAddress");
 	*len = 0;
 	*status = NDIS_STATUS_FAILURE;
 	if (RtlAnsiStringToUnicodeString(&key, &ansi, TRUE) != STATUS_SUCCESS)
@@ -1857,16 +1789,15 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisMRegisterInterrupt,7)
 	ndis_irq->shared = shared;
 	nt_spin_lock_init(&ndis_irq->lock);
 	wnd->ndis_irq = ndis_irq;
-
 	tasklet_init(&wnd->irq_tasklet, ndis_irq_handler, (unsigned long)wnd);
 	if (request_irq(vector, ndis_isr, req_isr ? SA_SHIRQ : 0,
-			"ndiswrapper", wnd)) {
-		printk(KERN_WARNING "%s: request for irq %d failed\n",
+			wnd->net_dev->name, wnd)) {
+		printk(KERN_WARNING "%s: request for IRQ %d failed\n",
 		       DRIVER_NAME, vector);
 		TRACEEXIT1(return NDIS_STATUS_RESOURCES);
 	}
 	ndis_irq->enabled = 1;
-	printk(KERN_INFO "%s: using irq %d\n", DRIVER_NAME, vector);
+	printk(KERN_INFO "%s: using IRQ %d\n", DRIVER_NAME, vector);
 	TRACEEXIT1(return NDIS_STATUS_SUCCESS);
 }
 
@@ -1974,8 +1905,7 @@ wstdcall void WIN_FUNC(NdisMIndicateStatus,4)
 			}
 
 			end = (u8 *)buf + len;
-			DBGTRACE2("PMKID_CANDIDATE_LIST ver %ld "
-				  "num_cand %ld",
+			DBGTRACE2("PMKID_CANDIDATE_LIST ver %ld num_cand %ld",
 				  cand->version, cand->num_candidates);
 			for (i = 0; i < cand->num_candidates; i++) {
 #if WIRELESS_EXT > 17
@@ -2071,10 +2001,9 @@ wstdcall void NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 	ndis_buffer *buffer;
 	struct ndis_packet *packet;
 	struct sk_buff *skb;
-	int i, length, total_length;
+	UINT i, length, total_length;
 	struct ndis_packet_oob_data *oob_data;
 	void *virt;
-	struct ndis_tcp_ip_checksum_packet_info *rx_csum_info;
 
 	TRACEENTER3("%p, %d", nmb, nr_packets);
 	wnd = nmb->wnd;
@@ -2084,6 +2013,7 @@ wstdcall void NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 			WARNING("empty packet ignored");
 			continue;
 		}
+		wnd->net_dev->last_rx = jiffies;
 		/* get total number of bytes in packet */
 		NdisGetFirstBufferFromPacketSafe(packet, &buffer, &virt,
 						 &length, &total_length,
@@ -2092,25 +2022,15 @@ wstdcall void NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 		oob_data = NDIS_PACKET_OOB_DATA(packet);
 		skb = dev_alloc_skb(total_length);
 		if (skb) {
-			skb->dev = wnd->net_dev;
 			while (buffer) {
-				virt = MmGetSystemAddressForMdl(buffer);
-				length = MmGetMdlByteCount(buffer);
-				memcpy(skb_put(skb, length), virt, length);
+				memcpy_skb(skb, MmGetSystemAddressForMdl(buffer),
+					   MmGetMdlByteCount(buffer));
 				buffer = buffer->next;
 			}
+			skb->dev = wnd->net_dev;
 			skb->protocol = eth_type_trans(skb, wnd->net_dev);
 			pre_atomic_add(wnd->stats.rx_bytes, total_length);
 			atomic_inc_var(wnd->stats.rx_packets);
-			rx_csum_info =
-				oob_data->extension.info[TcpIpChecksumPacketInfo];
-			if (wnd->rx_csum.ip_csum && rx_csum_info &&
-			    (rx_csum_info->rx.tcp_succeeded ||
-			     rx_csum_info->rx.ip_succeeded ||
-			     rx_csum_info->rx.udp_succeeded)) {
-				skb->ip_summed = CHECKSUM_HW;
-				skb->csum = rx_csum_info->value;
-			}
 			netif_rx(skb);
 		} else {
 			WARNING("couldn't allocate skb; packet dropped");
@@ -2130,7 +2050,7 @@ wstdcall void NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 		if (oob_data->status == NDIS_STATUS_RESOURCES)
 			continue;
 
-		assert (oob_data->status == NDIS_STATUS_SUCCESS);
+		assert(oob_data->status == NDIS_STATUS_SUCCESS);
 		/* deserialized driver doesn't check the status upon
 		 * return from this function; we need to call
 		 * MiniportReturnPacket later for this packet. Calling
@@ -2205,10 +2125,9 @@ wstdcall void EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx
 	KIRQL irql;
 	struct ndis_packet_oob_data *oob_data;
 
-	TRACEENTER3("nmb = %p, rx_ctx = %p, buf = %p, size = %d, "
-		    "buf = %p, size = %d, packet = %d",
-		    nmb, rx_ctx, header, header_size, look_ahead,
-		    look_ahead_size, packet_size);
+	TRACEENTER3("nmb = %p, rx_ctx = %p, buf = %p, size = %d, buf = %p, "
+		    "size = %d, packet = %d", nmb, rx_ctx, header, header_size,
+		    look_ahead, look_ahead_size, packet_size);
 
 	wnd = nmb->wnd;
 	DBGTRACE3("wnd = %p", wnd);
@@ -2216,6 +2135,7 @@ wstdcall void EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx
 		ERROR("nmb is NULL");
 		TRACEEXIT3(return);
 	}
+	wnd->net_dev->last_rx = jiffies;
 
 	if (look_ahead_size < packet_size) {
 		struct ndis_packet *packet;
@@ -2241,16 +2161,13 @@ wstdcall void EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx
 			skb = dev_alloc_skb(header_size + look_ahead_size +
 					    bytes_txed);
 			if (skb) {
-				memcpy(skb_put(skb, header_size), header,
-				       header_size);
-				memcpy(skb_put(skb, look_ahead_size),
-				       look_ahead, look_ahead_size);
+				memcpy_skb(skb, header, header_size);
+				memcpy_skb(skb, look_ahead, look_ahead_size);
 				buffer = packet->private.buffer_head;
 				while (buffer) {
-					int length = MmGetMdlByteCount(buffer);
-					memcpy(skb_put(skb, length),
-					       MmGetSystemAddressForMdl(buffer),
-					       length);
+					memcpy_skb(skb,
+						   MmGetSystemAddressForMdl(buffer),
+						   MmGetMdlByteCount(buffer));
 					buffer = buffer->next;
 				}
 				skb_size = header_size+look_ahead_size +
@@ -2283,9 +2200,8 @@ wstdcall void EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx
 		skb_size = header_size + packet_size;
 		skb = dev_alloc_skb(skb_size);
 		if (skb) {
-			memcpy(skb_put(skb, header_size), header, header_size);
-			memcpy(skb_put(skb, packet_size), look_ahead,
-			       packet_size);
+			memcpy_skb(skb, header, header_size);
+			memcpy_skb(skb, look_ahead, packet_size);
 		}
 	}
 
@@ -2313,20 +2229,17 @@ wstdcall void NdisMTransferDataComplete(struct ndis_miniport_block *nmb,
 	unsigned int skb_size;
 	struct ndis_packet_oob_data *oob_data;
 	ndis_buffer *buffer;
-	struct ndis_tcp_ip_checksum_packet_info *rx_csum_info;
 
 	TRACEENTER3("wnd = %p, packet = %p, bytes_txed = %d",
 		    wnd, packet, bytes_txed);
-
 	if (!packet) {
 		WARNING("illegal packet");
 		TRACEEXIT3(return);
 	}
-
+	wnd->net_dev->last_rx = jiffies;
 	oob_data = NDIS_PACKET_OOB_DATA(packet);
 	skb_size = sizeof(oob_data->header) + oob_data->look_ahead_size +
 		bytes_txed;
-
 	skb = dev_alloc_skb(skb_size);
 	if (!skb) {
 		kfree(oob_data->look_ahead);
@@ -2335,31 +2248,20 @@ wstdcall void NdisMTransferDataComplete(struct ndis_miniport_block *nmb,
 		atomic_inc_var(wnd->stats.rx_dropped);
 		TRACEEXIT3(return);
 	}
-
-	skb->dev = wnd->net_dev;
-	memcpy(skb_put(skb, sizeof(oob_data->header)), oob_data->header,
-	       sizeof(oob_data->header));
-	memcpy(skb_put(skb, oob_data->look_ahead_size), oob_data->look_ahead,
-	       oob_data->look_ahead_size);
+	memcpy_skb(skb, oob_data->header, sizeof(oob_data->header));
+	memcpy_skb(skb, oob_data->look_ahead, oob_data->look_ahead_size);
 	buffer = packet->private.buffer_head;
 	while (buffer) {
-		memcpy(skb_put(skb, MmGetMdlByteCount(buffer)),
-		       MmGetSystemAddressForMdl(buffer),
-		       MmGetMdlByteCount(buffer));
+		memcpy_skb(skb, MmGetSystemAddressForMdl(buffer),
+			   MmGetMdlByteCount(buffer));
 		buffer = buffer->next;
 	}
 	kfree(oob_data->look_ahead);
 	NdisFreePacket(packet);
+	skb->dev = wnd->net_dev;
 	skb->protocol = eth_type_trans(skb, wnd->net_dev);
 	pre_atomic_add(wnd->stats.rx_bytes, skb_size);
 	atomic_inc_var(wnd->stats.rx_packets);
-	rx_csum_info = oob_data->extension.info[TcpIpChecksumPacketInfo];
-	if (wnd->rx_csum.ip_csum && rx_csum_info &&
-	    (rx_csum_info->rx.tcp_succeeded || rx_csum_info->rx.ip_succeeded ||
-	     rx_csum_info->rx.udp_succeeded)) {
-		skb->ip_summed = CHECKSUM_HW;
-		skb->csum = rx_csum_info->value;
-	}
 	netif_rx(skb);
 }
 
@@ -2537,6 +2439,33 @@ wstdcall void NdisMResetComplete(struct ndis_miniport_block *nmb,
 	TRACEEXIT3(return);
 }
 
+static void ndis_worker(void *dummy)
+{
+	KIRQL irql;
+	struct ndis_work_entry *ndis_work_entry;
+	struct nt_list *ent;
+	struct ndis_work_item *ndis_work_item;
+
+	WORKENTER("");
+	while (1) {
+		irql = nt_spin_lock_irql(&ndis_work_list_lock, DISPATCH_LEVEL);
+		ent = RemoveHeadList(&ndis_worker_list);
+		nt_spin_unlock_irql(&ndis_work_list_lock, irql);
+		if (!ent)
+			break;
+		ndis_work_entry = container_of(ent, struct ndis_work_entry,
+					       list);
+		ndis_work_item = ndis_work_entry->ndis_work_item;
+		WORKTRACE("%p: %p, %p", ndis_work_item,
+			  ndis_work_item->func, ndis_work_item->ctx);
+		LIN2WIN2(ndis_work_item->func, ndis_work_item,
+			 ndis_work_item->ctx);
+		WORKTRACE("%p done", ndis_work_item);
+		kfree(ndis_work_entry);
+	}
+	WORKEXIT(return);
+}
+
 wstdcall NDIS_STATUS WIN_FUNC(NdisScheduleWorkItem,1)
 	(struct ndis_work_item *ndis_work_item)
 {
@@ -2544,7 +2473,7 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisScheduleWorkItem,1)
 	KIRQL irql;
 
 	TRACEENTER3("%p", ndis_work_item);
-	ndis_work_entry = kmalloc(sizeof(*ndis_work_entry), GFP_ATOMIC);
+	ndis_work_entry = kmalloc(sizeof(*ndis_work_entry), gfp_irql());
 	if (!ndis_work_entry)
 		BUG();
 	ndis_work_entry->ndis_work_item = ndis_work_item;
@@ -2554,7 +2483,6 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisScheduleWorkItem,1)
 	WORKTRACE("scheduling %p", ndis_work_item);
 	schedule_ndis_work(&ndis_work);
 	TRACEEXIT3(return NDIS_STATUS_SUCCESS);
-
 }
 
 wstdcall void WIN_FUNC(NdisMGetDeviceProperty,6)
@@ -2586,7 +2514,7 @@ wstdcall NDIS_STATUS WIN_FUNC(NdisMQueryAdapterInstanceName,2)
 	struct wrap_ndis_device *wnd = nmb->wnd;
 	struct ansi_string ansi;
 
-	if (wrap_is_pci_bus(wnd->wd->dev_bus_type))
+	if (wrap_is_pci_bus(wnd->wd->dev_bus))
 		RtlInitAnsiString(&ansi, "PCI Ethernet Adapter");
 	else
 		RtlInitAnsiString(&ansi, "USB Ethernet Adapter");
@@ -2644,6 +2572,22 @@ wstdcall void WIN_FUNC(NdisMCoRequestComplete,3)
 	TRACEEXIT3(return);
 }
 
+wstdcall NDIS_STATUS WIN_FUNC(NdisMSetMiniportSecondary,2)
+	(struct ndis_miniport_block *nmb2, struct ndis_miniport_block *nmb1)
+{
+	TRACEENTER3("%p, %p", nmb1, nmb2);
+	TODO();
+	TRACEEXIT3(return NDIS_STATUS_SUCCESS);
+}
+
+wstdcall NDIS_STATUS WIN_FUNC(NdisMPromoteMiniport,1)
+	(struct ndis_miniport_block *nmb)
+{
+	TRACEENTER3("%p", nmb);
+	TODO();
+	TRACEEXIT3(return NDIS_STATUS_SUCCESS);
+}
+
 wstdcall void WIN_FUNC(NdisMCoActivateVcComplete,3)
 	(NDIS_STATUS status, void *handle, void *params)
 {
