@@ -421,11 +421,12 @@ struct ndis_rw_lock {
 	union ndis_rw_lock_refcount ref_count[MAXIMUM_PROCESSORS];
 };
 
+struct ndis_work_item;
+typedef void (*NDIS_PROC)(struct ndis_work_item *, void *) wstdcall;
+
 struct ndis_work_item {
 	void *ctx;
-	/* this should be NDIS_PROC, but we masquerade it as
-	 * NTOS_WORK_FUNC so we can use wrap_worker */
-	NTOS_WORK_FUNC func wstdcall;
+	NDIS_PROC func;
 	UCHAR reserved[8 * sizeof(void *)];
 };
 
@@ -799,24 +800,25 @@ struct wrap_ndis_device {
 
 	struct work_struct tx_work;
 	struct ndis_packet *tx_ring[TX_RING_SIZE];
-	unsigned int tx_ring_start;
-	unsigned int tx_ring_end;
-	unsigned char is_tx_ring_full;
+	u8 tx_ring_start;
+	u8 tx_ring_end;
+	u8 is_tx_ring_full;
 	struct semaphore tx_ring_mutex;
 	unsigned int max_tx_packets;
 	u8 tx_ok;
 	struct semaphore ndis_comm_mutex;
 	wait_queue_head_t ndis_comm_wq;
-	int ndis_comm_done;
-	int ndis_comm_status;
+	s8 ndis_comm_done;
+	NDIS_STATUS ndis_comm_status;
 	ULONG packet_filter;
 
 	BOOLEAN use_sg_dma;
-	int dma_map_count;
-	dma_addr_t *dma_map_addr;
+	u8 tx_dma_count;
+	dma_addr_t *tx_dma_addr;
 
 	int hangcheck_interval;
 	struct timer_list hangcheck_timer;
+	int stats_interval;
 	struct timer_list stats_timer;
 	unsigned long scan_timestamp;
 	struct encr_info encr_info;
