@@ -33,7 +33,7 @@
 			(sizeof(PFN_NUMBER) * CACHE_MDL_PAGES))
 struct wrap_mdl {
 	struct nt_list list;
-	char mdl[CACHE_MDL_SIZE];
+	struct mdl mdl[0];
 };
 
 struct thread_event_waitq {
@@ -78,7 +78,6 @@ struct bus_driver {
 };
 
 static struct nt_list bus_driver_list;
-static void del_bus_drivers(void);
 
 static struct work_struct ntos_work_item_work;
 static struct nt_list ntos_work_item_list;
@@ -87,7 +86,6 @@ static void ntos_work_item_worker(void *data);
 
 NT_SPIN_LOCK irp_cancel_lock;
 
-extern NT_SPIN_LOCK loader_lock;
 extern struct nt_list wrap_drivers;
 static struct nt_list wrap_timer_list;
 NT_SPIN_LOCK timer_lock;
@@ -103,7 +101,6 @@ static void update_user_shared_data_proc(unsigned long data);
 #endif
 
 static int add_bus_driver(const char *name);
-static void free_all_objects(void);
 static BOOLEAN queue_kdpc(struct kdpc *kdpc);
 
 WIN_SYMBOL_MAP("KeTickCount", &jiffies)
@@ -158,7 +155,8 @@ int ntoskernel_init(void)
 		ntoskernel_exit();
 		return -ENOMEM;
 	}
-	mdl_cache = kmem_cache_create("wrap_mdl", sizeof(struct wrap_mdl),
+	mdl_cache = kmem_cache_create("wrap_mdl",
+				      sizeof(struct wrap_mdl) + CACHE_MDL_SIZE,
 				      0, 0, NULL, NULL);
 	DBGTRACE2("%p", mdl_cache);
 	if (!mdl_cache) {
@@ -200,6 +198,10 @@ void ntoskernel_exit(void)
 	KIRQL irql;
 
 	TRACEENTER2("");
+
+#ifdef KDPC_TASKLET
+	tasklet_kill(&kdpc_work);
+#endif
 	/* free kernel (Ke) timers */
 	DBGTRACE2("freeing timers");
 	while (1) {
@@ -221,19 +223,16 @@ void ntoskernel_exit(void)
 	DBGTRACE2("freeing MDLs");
 	if (mdl_cache) {
 		irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
-		if (!IsListEmpty(&wrap_mdl_list)) {
+		if (!IsListEmpty(&wrap_mdl_list))
 			ERROR("Windows driver didn't free all MDLs; "
 			      "freeing them now");
-			while ((cur = RemoveHeadList(&wrap_mdl_list))) {
-				struct wrap_mdl *p;
-				struct mdl *mdl;
-				p = container_of(cur, struct wrap_mdl, list);
-				mdl = (struct mdl *)p->mdl;
-				if (mdl->flags & MDL_CACHE_ALLOCATED)
-					kmem_cache_free(mdl_cache, p);
-				else
-					kfree(p);
-			}
+		while ((cur = RemoveHeadList(&wrap_mdl_list))) {
+			struct wrap_mdl *wrap_mdl;
+			wrap_mdl = container_of(cur, struct wrap_mdl, list);
+			if (wrap_mdl->mdl->flags & MDL_CACHE_ALLOCATED)
+				kmem_cache_free(mdl_cache, wrap_mdl);
+			else
+				kfree(wrap_mdl);
 		}
 		nt_spin_unlock_irql(&ntoskernel_lock, irql);
 		kmem_cache_destroy(mdl_cache);
@@ -265,8 +264,26 @@ void ntoskernel_exit(void)
 	}
 	nt_spin_unlock_irql(&dispatcher_lock, irql);
 
-	del_bus_drivers();
-	free_all_objects();
+	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
+	while ((cur = RemoveHeadList(&bus_driver_list))) {
+		struct bus_driver *bus_driver;
+		bus_driver = container_of(cur, struct bus_driver, list);
+		/* TODO: make sure all all drivers are shutdown/removed */
+		kfree(bus_driver);
+	}
+	nt_spin_unlock_irql(&ntoskernel_lock, irql);
+
+	TRACEENTER2("freeing objects");
+	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
+	while ((cur = RemoveHeadList(&object_list))) {
+		struct common_object_header *hdr;
+		hdr = container_of(cur, struct common_object_header, list);
+		WARNING("object %p type %d was not freed, freeing it now",
+			HEADER_TO_OBJECT(hdr), hdr->type);
+		ExFreePool(hdr);
+	}
+	nt_spin_unlock_irql(&ntoskernel_lock, irql);
+
 #if defined(CONFIG_X86_64)
 	del_timer_sync(&shared_data_timer);
 #endif
@@ -347,29 +364,6 @@ void free_object(void *object)
 	ExFreePool(hdr);
 }
 
-static void free_all_objects(void)
-{
-	struct nt_list *cur;
-
-	TRACEENTER2("freeing objects");
-	/* delete all objects */
-	while (1) {
-		struct common_object_header *hdr;
-		KIRQL irql;
-
-		irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
-		cur = RemoveHeadList(&object_list);
-		nt_spin_unlock_irql(&ntoskernel_lock, irql);
-		if (!cur)
-			break;
-		hdr = container_of(cur, struct common_object_header, list);
-		WARNING("object %p type %d was not freed, freeing it now",
-			HEADER_TO_OBJECT(hdr), hdr->type);
-		ExFreePool(hdr);
-	}
-	TRACEEXIT2(return);
-}
-
 static int add_bus_driver(const char *name)
 {
 	struct bus_driver *bus_driver;
@@ -389,21 +383,6 @@ static int add_bus_driver(const char *name)
 	return STATUS_SUCCESS;
 }
 
-static void del_bus_drivers(void)
-{
-	struct nt_list *ent;
-	KIRQL irql;
-
-	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
-	while ((ent = RemoveHeadList(&bus_driver_list))) {
-		struct bus_driver *bus_driver;
-		bus_driver = container_of(ent, struct bus_driver, list);
-		/* TODO: make sure all all drivers are shutdown/removed */
-		kfree(bus_driver);
-	}
-	nt_spin_unlock_irql(&ntoskernel_lock, irql);
-}
-
 struct driver_object *find_bus_driver(const char *name)
 {
 	struct bus_driver *bus_driver;
@@ -611,11 +590,11 @@ wfastcall void WIN_FUNC(ExInterlockedAddLargeStatistic,2)
 		"   movl %%edx, %%ecx\n\t"
 		"   addl %%eax, %%ebx\n\t"
 		"   adcl $0, %%ecx\n\t"
-		    LOCK_PREFIX "cmpxchg8b (%0)\n\t"
+		    LOCK_PREFIX "cmpxchg8b %0\n\t"
 		"   jnz 1b\n\t"
-		: "+r" (plint)
-		: "m" (n), "a" (ll_low(*plint)), "d" (ll_high(*plint))
-		: "ebx", "ecx", "cc");
+		: "+m" (*plint)
+		: "m" (n), "A" (*plint)
+		: "ebx", "ecx");
 #endif
 	restore_local_irq(flags);
 }
@@ -1262,6 +1241,7 @@ static void wakeup_threads(struct dispatcher_header *dh)
 	struct nt_list *cur, *next;
 	struct wait_block *wb = NULL;
 
+	EVENTENTER("%p", dh);
 	nt_list_for_each_safe(cur, next, &dh->wait_blocks) {
 		wb = container_of(cur, struct wait_block, list);
 		EVENTTRACE("%p: wait block: %p, thread: %p",
@@ -1344,7 +1324,7 @@ wstdcall NTSTATUS WIN_FUNC(KeWaitForMultipleObjects,8)
 	KIRQL irql;
 
 	thread = current;
-	EVENTTRACE("thread: %p count = %d, type: %d, reason = %u, "
+	EVENTENTER("thread: %p count = %d, type: %d, reason = %u, "
 		   "waitmode = %u, alertable = %u, timeout = %p", thread,
 		   count, wait_type, wait_reason, wait_mode, alertable,
 		   timeout);
@@ -1650,7 +1630,11 @@ wstdcall LONG WIN_FUNC(KeReleaseSemaphore,4)
 	assert(ret >= 0);
 	if (semaphore->dh.signal_state + adjustment <= semaphore->limit)
 		semaphore->dh.signal_state += adjustment;
-	/* else raise exception */
+	else {
+		WARNING("releasing %d over limit %d", adjustment,
+			semaphore->limit);
+		semaphore->dh.signal_state = semaphore->limit;
+	}
 	if (semaphore->dh.signal_state > 0)
 		wakeup_threads(&semaphore->dh);
 	nt_spin_unlock_irql(&dispatcher_lock, irql);
@@ -1692,6 +1676,8 @@ wstdcall KPRIORITY WIN_FUNC(KeQueryPriorityThread,1)
 	KPRIORITY prio;
 
 	EVENTENTER("task: %p", task);
+	return LOW_REALTIME_PRIORITY;
+
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 	prio = 1;
 #else
@@ -1752,6 +1738,8 @@ wstdcall KPRIORITY WIN_FUNC(KeSetPriorityThread,2)
 
 	TRACEENTER3("task: %p, priority = %u", task, priority);
 
+	return LOW_REALTIME_PRIORITY;
+
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 	/* FIXME: is there a way to set kernel thread prio on 2.4? */
 	old_prio = LOW_PRIORITY;
@@ -1992,9 +1980,10 @@ wstdcall void *WIN_FUNC(MmAllocateContiguousMemorySpecifyCache,5)
 	 PHYSICAL_ADDRESS boundary, enum memory_caching_type cache_type)
 {
 	void *addr;
-	DBGTRACE2("%lu, %Lu, %Lu, %Lu, %d", size, lowest, highest, boundary,
-		  cache_type);
-	addr = ExAllocatePoolWithTag(NonPagedPool, size, 0);
+	size_t page_length = ((size + PAGE_SIZE - 1) / PAGE_SIZE) * PAGE_SIZE;
+	DBGTRACE2("%lu, %lu, %Lu, %Lu, %Lu, %d", size, page_length,
+		  lowest, highest, boundary, cache_type);
+	addr = ExAllocatePoolWithTag(NonPagedPool, page_length, 0);
 	DBGTRACE2("%p", addr);
 	return addr;
 }
@@ -2051,7 +2040,7 @@ wstdcall ULONG WIN_FUNC(MmSizeOfMdl,2)
 struct mdl *allocate_init_mdl(void *virt, ULONG length)
 {
 	struct wrap_mdl *wrap_mdl;
-	struct mdl *mdl = NULL;
+	struct mdl *mdl;
 	int mdl_size = MmSizeOfMdl(virt, length);
 	KIRQL irql;
 
@@ -2062,7 +2051,7 @@ struct mdl *allocate_init_mdl(void *virt, ULONG length)
 		irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
 		InsertHeadList(&wrap_mdl_list, &wrap_mdl->list);
 		nt_spin_unlock_irql(&ntoskernel_lock, irql);
-		mdl = (struct mdl *)wrap_mdl->mdl;
+		mdl = wrap_mdl->mdl;
 		DBGTRACE3("allocated mdl from cache: %p(%p), %p(%d)",
 			  wrap_mdl, mdl, virt, length);
 		memset(mdl, 0, CACHE_MDL_SIZE);
@@ -2072,11 +2061,10 @@ struct mdl *allocate_init_mdl(void *virt, ULONG length)
 		mdl->flags = MDL_CACHE_ALLOCATED;
 	} else {
 		wrap_mdl =
-			kmalloc(sizeof(*wrap_mdl) + mdl_size - CACHE_MDL_SIZE,
-				gfp_irql());
+			kmalloc(sizeof(*wrap_mdl) + mdl_size, gfp_irql());
 		if (!wrap_mdl)
 			return NULL;
-		mdl = (struct mdl *)wrap_mdl->mdl;
+		mdl = wrap_mdl->mdl;
 		DBGTRACE3("allocated mdl from memory: %p(%p), %p(%d)",
 			  wrap_mdl, mdl, virt, length);
 		irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
@@ -2096,14 +2084,13 @@ void free_mdl(struct mdl *mdl)
 	 * with IoFreeMdl (e.g., 64-bit Broadcom). Since we need to
 	 * treat buffers allocated with Ndis calls differently, we
 	 * must call NdisFreeBuffer if it is allocated with Ndis
-	 * function. We set 'process' field in Ndis functions. */
+	 * function. We set 'pool' field in Ndis functions. */
 	if (!mdl)
 		return;
-	if (mdl->process)
+	if (mdl->pool)
 		NdisFreeBuffer(mdl);
 	else {
-		struct wrap_mdl *wrap_mdl;
-		wrap_mdl = (struct wrap_mdl *)
+		struct wrap_mdl *wrap_mdl = (struct wrap_mdl *)
 			((char *)mdl - offsetof(struct wrap_mdl, mdl));
 		irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
 		RemoveEntryList(&wrap_mdl->list);
