@@ -16,8 +16,6 @@
 #ifndef _NTOSKERNEL_H_
 #define _NTOSKERNEL_H_
 
-#define UTILS_VERSION "1.8"
-
 #include <linux/types.h>
 #include <linux/timer.h>
 #include <linux/time.h>
@@ -127,7 +125,7 @@ struct work_struct {
 struct workqueue_struct *create_singlethread_workqueue(const char *name);
 void destroy_workqueue(struct workqueue_struct *workq);
 void queue_work(struct workqueue_struct *workq,
-		struct work_struct *work_struct) fastcall;
+		struct work_struct *work_struct) wfastcall;
 void cancel_delayed_work(struct work_struct *work_struct);
 
 #include <linux/smp_lock.h>
@@ -334,10 +332,14 @@ typedef u32 pm_message_t;
 #define netdev_priv(dev)  ((dev)->priv)
 #endif
 
+#define memcpy_skb(skb, from, length)			\
+	memcpy(skb_put(skb, length), from, length)
+
 #include "ndiswrapper.h"
 #include "pe_linker.h"
 #include "wrapmem.h"
 #include "lin2win.h"
+#include "loader.h"
 
 #ifdef CONFIG_X86_64
 #define get_sp(sp) __asm__ __volatile__("mov %%rsp, %0\n\t" : "=m"(sp))
@@ -380,7 +382,7 @@ typedef u32 pm_message_t;
 	  (((s64)HZ * ((sys_time) - ticks_1601())) / TICKSPERSEC)))
 
 #define MSEC_TO_HZ(ms) ((ms) * HZ / 1000)
-#define USEC_TO_HZ(ms) ((us) * HZ / 1000000)
+#define USEC_TO_HZ(us) ((us) * HZ / 1000000)
 
 extern u64 wrap_ticks_to_boot;
 
@@ -460,7 +462,6 @@ struct wrap_device_setting {
 };
 
 struct wrap_bin_file {
-	char driver_name[MAX_DRIVER_NAME_LEN];
 	char name[MAX_DRIVER_NAME_LEN];
 	int size;
 	void *data;
@@ -485,15 +486,15 @@ struct usbd_pipe_information;
 
 struct wrap_device {
 	/* first part is (de)initialized once by loader */
-	int dev_bus_type;
+	struct nt_list list;
+	int dev_bus;
 	int vendor;
 	int device;
 	int subvendor;
 	int subdevice;
-	struct wrap_driver *driver;
-	/* we need driver_name before driver is loaded */
-	char driver_name[MAX_DRIVER_NAME_LEN];
 	char conf_file_name[MAX_DRIVER_NAME_LEN];
+	char driver_name[MAX_DRIVER_NAME_LEN];
+	struct wrap_driver *driver;
 	struct nt_list settings;
 
 	/* rest should be (de)initialized when a device is
@@ -520,19 +521,20 @@ struct wrap_device {
 	BOOLEAN surprise_removed;
 };
 
-#define wrap_is_pci_bus(dev_bus_type)				\
-	(WRAP_BUS_TYPE(dev_bus_type) == WRAP_PCI_BUS)
+#define wrap_is_pci_bus(dev_bus)			\
+	(WRAP_BUS(dev_bus) == WRAP_PCI_BUS ||		\
+	 WRAP_BUS(dev_bus) == WRAP_PCMCIA_BUS)
 #ifdef CONFIG_USB
 /* earlier versions of ndiswrapper used 0 as USB_BUS */
-#define wrap_is_usb_bus(dev_bus_type)				\
-	(WRAP_BUS_TYPE(dev_bus_type) == WRAP_USB_BUS ||		\
-	 WRAP_BUS_TYPE(dev_bus_type) == 0)
+#define wrap_is_usb_bus(dev_bus)			\
+	(WRAP_BUS(dev_bus) == WRAP_USB_BUS ||		\
+	 WRAP_BUS(dev_bus) == WRAP_INTERNAL_BUS)
 #else
-#define wrap_is_usb_bus(dev_bus_type) 0
+#define wrap_is_usb_bus(dev_bus) 0
 #endif
-#define wrap_is_bluetooth_device(dev_bus_type)				\
-	(WRAP_DEVICE_TYPE(dev_bus_type) == WRAP_BLUETOOTH_DEVICE1 ||	\
-	 WRAP_DEVICE_TYPE(dev_bus_type) == WRAP_BLUETOOTH_DEVICE2)
+#define wrap_is_bluetooth_device(dev_bus)			\
+	(WRAP_DEVICE(dev_bus) == WRAP_BLUETOOTH_DEVICE1 ||	\
+	 WRAP_DEVICE(dev_bus) == WRAP_BLUETOOTH_DEVICE2)
 
 /* Some drivers use worker entries to complete functions called from
  * within other worker threads. So we should have separate workqueues
@@ -581,8 +583,10 @@ int usb_init_device(struct wrap_device *wd);
 void usb_exit_device(struct wrap_device *wd);
 void usb_cancel_pending_urbs(void);
 
-int misc_funcs_init(void);
-void misc_funcs_exit(void);
+int crt_init(void);
+void crt_exit(void);
+int rtl_init(void);
+void rtl_exit(void);
 int wrap_procfs_init(void);
 void wrap_procfs_remove(void);
 int stricmp(const char *s1, const char *s2);
@@ -752,9 +756,10 @@ static inline KIRQL current_irql(void)
 {
 	if (in_irq() || irqs_disabled())
 		TRACEEXIT5(return DEVICE_LEVEL);
-	if (in_atomic())
+	else if (in_atomic())
 		TRACEEXIT5(return DISPATCH_LEVEL);
-	TRACEEXIT5(return PASSIVE_LEVEL);
+	else
+		TRACEEXIT5(return PASSIVE_LEVEL);
 }
 
 static inline KIRQL raise_irql(KIRQL newirql)
@@ -773,8 +778,10 @@ static inline void lower_irql(KIRQL oldirql)
 {
 	KIRQL irql = current_irql();
 	DBGTRACE5("%d, %d", irql, oldirql);
-	if (irql < oldirql)
-		ERROR("invalid irql: %d < %d", irql, oldirql);
+	DBG_BLOCK(2) {
+		if (irql < oldirql)
+			ERROR("invalid irql: %d < %d", irql, oldirql);
+	}
 	if (oldirql < DISPATCH_LEVEL && irql == DISPATCH_LEVEL) {
 		preempt_enable();
 		local_bh_enable();
@@ -815,9 +822,8 @@ static inline void nt_spin_lock(volatile NT_SPIN_LOCK *lock)
 		"  jne 2b\n\t"
 		"  jmp 1b\n"
 		"3:\n\t"
-		: "=m" (*lock)
-		: "r" (NT_SPIN_LOCK_LOCKED), "i" (NT_SPIN_LOCK_UNLOCKED),
-		  "m" (*lock));
+		: "+m" (*lock)
+		: "r" (NT_SPIN_LOCK_LOCKED), "i" (NT_SPIN_LOCK_UNLOCKED));
 }
 
 static inline void nt_spin_unlock(volatile NT_SPIN_LOCK *lock)
@@ -875,19 +881,19 @@ do {									\
 		if (size == 1)					\
 			__asm__ __volatile__(			\
 				LOCK_PREFIX oper "b %b0\n\t"	\
-				: "=m" (var) : "m" (var));	\
+				: "+m" (var));			\
 		else if (size == 2)				\
 			__asm__ __volatile__(			\
 				LOCK_PREFIX oper "w %w0\n\t"	\
-				: "=m" (var) : "m" (var));	\
+				: "+m" (var));			\
 		else if (size == 4)				\
 			__asm__ __volatile__(			\
 				LOCK_PREFIX oper "l %0\n\t"	\
-				: "=m" (var) : "m" (var));	\
+				: "+m" (var));			\
 		else if (size == 8)				\
 			__asm__ __volatile__(			\
 				LOCK_PREFIX oper "q %q0\n\t"	\
-				: "=m" (var) : "m" (var));	\
+				: "+m" (var));			\
 		else						\
 			ERROR("invalid size %d", (int)size);	\
 	} while (0)
@@ -905,21 +911,33 @@ do {									\
 	typeof(var) pre;					\
 	__asm__ __volatile__(					\
 		LOCK_PREFIX "xadd %0, %1\n\t"			\
-		: "=r"(pre), "=m"(var)				\
-		: "0"(i), "m" (var));				\
+		: "=r"(pre), "+m"(var)				\
+		: "0"(i));					\
 	pre;							\
 })
 
 #define post_atomic_add(var, i) (pre_atomic_add(var, i) + i)
 
+#define atomic_insert_list_head(head, oldhead, newhead)			\
+	do {								\
+		oldhead = head;						\
+	} while (cmpxchg(&(head), oldhead, newhead) != oldhead)
+
+#define atomic_remove_list_head(head, newhead)				\
+({									\
+	typeof(head) oldhead;						\
+	do {								\
+		oldhead = head;						\
+		if (!oldhead)						\
+			break;						\
+	} while (cmpxchg(&(head), oldhead, newhead) != oldhead);	\
+	oldhead;							\
+})
+
 static inline ULONG SPAN_PAGES(void *ptr, SIZE_T length)
 {
-	ULONG n;
-
-	n = ((((ULONG_PTR)ptr) & (PAGE_SIZE - 1)) + length + (PAGE_SIZE - 1))
-		>> PAGE_SHIFT;
-
-	return n;
+	return PAGE_ALIGN(((unsigned long)ptr & (PAGE_SIZE - 1)) + length)
+			  >> PAGE_SHIFT;
 }
 
 #ifdef CONFIG_X86_64
@@ -956,23 +974,18 @@ static inline struct nt_slist *PopEntrySList(nt_slist_header *head,
 
 #else
 
-#ifndef ll_low
-#define ll_low(x) *(((u32 *)&(x)) + 0)
-#define ll_high(x) *(((u32 *)&(x)) + 1)
-#endif
+#define u64_low_32(x) ((u32)x)
+#define u64_high_32(x) ((u32)(x >> 32))
 
-static inline u64 cmpxchg8b(volatile void *ptr, u64 old, u64 new)
+static inline u64 cmpxchg8b(volatile u64 *ptr, u64 old, u64 new)
 {
 	u64 prev;
 
 	__asm__ __volatile__(
 		"\n"
-		LOCK_PREFIX "cmpxchg8b (%0)\n"
-		: "+r" (ptr),
-		  "=a" (ll_low(prev)), "=d" (ll_high(prev))
-		: "a" (ll_low(old)), "d" (ll_high(old)),
-		  "b" (ll_low(new)), "c" (ll_high(new))
-		: "cc");
+		LOCK_PREFIX "cmpxchg8b %0\n"
+		: "+m" (*ptr), "=A" (prev)
+		: "A" (old), "b" (u64_low_32(new)), "c" (u64_high_32(new)));
 	return prev;
 }
 
