@@ -65,10 +65,8 @@ wstdcall void WIN_FUNC(IoInitializeIrp,3)
 	memset(irp, 0, size);
 	irp->size = size;
 	irp->stack_count = stack_count;
-	/* IoAllocateIrp allocates space for one more than requested
-	 * stack_count */
-	irp->current_location = stack_count + 1;
-	IoGetCurrentIrpStackLocation(irp) = IRP_SL(irp, stack_count + 1);
+	irp->current_location = stack_count;
+	IoGetCurrentIrpStackLocation(irp) = IRP_SL(irp, stack_count);
 	IOEXIT(return);
 }
 
@@ -94,10 +92,8 @@ wstdcall struct irp *WIN_FUNC(IoAllocateIrp,2)
 	int irp_size;
 
 	IOENTER("count: %d", stack_count);
-	/* driver need not allocate stack location for itself, but we
-	 * need to allocate space for it so that driver can set major
-	 * function etc. even if stack_count is 0 */
-	irp_size = IoSizeOfIrp(stack_count + 1);
+	stack_count++;
+	irp_size = IoSizeOfIrp(stack_count);
 	irp = kmalloc(irp_size, gfp_irql());
 	if (irp) {
 		IOTRACE("allocated irp %p", irp);
@@ -144,8 +140,8 @@ wstdcall void IoQueueThreadIrp(struct irp *irp)
 	thread = get_current_nt_thread();
 	if (thread) {
 		IOTRACE("thread: %p, task: %p", thread, thread->task);
-		irql = nt_spin_lock_irql(&thread->lock, DISPATCH_LEVEL);
 		irp->flags |= IRP_SYNCHRONOUS_API;
+		irql = nt_spin_lock_irql(&thread->lock, DISPATCH_LEVEL);
 		InsertTailList(&thread->irps, &irp->threads);
 		IoIrpThread(irp) = thread;
 		nt_spin_unlock_irql(&thread->lock, irql);
@@ -290,12 +286,10 @@ wstdcall struct irp *WIN_FUNC(IoBuildDeviceIoControlRequest,9)
 				       input_buf, input_buf_len);
 			irp->flags = IRP_BUFFERED_IO | IRP_DEALLOCATE_BUFFER;
 			if (output_buf)
-				irp->flags |= IRP_INPUT_OPERATION;
+				irp->flags = IRP_INPUT_OPERATION;
 			irp->user_buf = output_buf;
-		} else { // buf_len = 0
-			irp->flags = 0;
+		} else
 			irp->user_buf = NULL;
-		}
 		break;
 	case METHOD_IN_DIRECT:
 	case METHOD_OUT_DIRECT:
@@ -310,8 +304,7 @@ wstdcall struct irp *WIN_FUNC(IoBuildDeviceIoControlRequest,9)
 			memcpy(irp->associated_irp.system_buffer,
 			       input_buf, input_buf_len);
 			irp->flags = IRP_BUFFERED_IO | IRP_DEALLOCATE_BUFFER;
-		} else
-			irp->flags = 0;
+		}
 		/* TODO: we are supposed to setup MDL, but USB layer
 		 * doesn't use MDLs. Moreover, USB layer mirrors
 		 * non-DMAable buffers, so no need to allocate
@@ -325,8 +318,7 @@ wstdcall struct irp *WIN_FUNC(IoBuildDeviceIoControlRequest,9)
 				IOEXIT(return NULL);
 			}
 			irp->flags = IRP_BUFFERED_IO | IRP_DEALLOCATE_BUFFER;
-		} else
-			irp->flags = 0;
+		}
 		break;
 	case METHOD_NEITHER:
 		irp->user_buf = output_buf;
@@ -352,6 +344,12 @@ wfastcall NTSTATUS WIN_FUNC(IofCallDriver,2)
 
 	IoSetNextIrpStackLocation(irp);
 	DUMP_IRP(irp);
+#ifdef IO_DEBUG
+	if (irp->current_location < 0) {
+		ERROR("invalid irp: %p, %d", irp, irp->current_location);
+		return STATUS_INVALID_PARAMETER;
+	}
+#endif
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 	drv_obj = dev_obj->drv_obj;
 	irp_sl->dev_obj = dev_obj;
@@ -372,10 +370,6 @@ wfastcall NTSTATUS WIN_FUNC(IofCallDriver,2)
 wfastcall void WIN_FUNC(IofCompleteRequest,2)
 	(struct irp *irp, CHAR prio_boost)
 {
-	NTSTATUS status;
-	struct io_stack_location *irp_sl;
-	struct mdl *mdl;
-
 #ifdef IO_DEBUG
 	DUMP_IRP(irp);
 	if (irp->io_status.status == STATUS_PENDING) {
@@ -387,10 +381,12 @@ wfastcall void WIN_FUNC(IofCompleteRequest,2)
 		return;
 	}
 #endif
-	for (irp_sl = IoGetCurrentIrpStackLocation(irp);
-	     irp->current_location <= irp->stack_count; irp_sl++) {
+	while (irp->current_location < irp->stack_count) {
+		struct io_stack_location *irp_sl;
 		struct device_object *dev_obj;
+		NTSTATUS status;
 
+		irp_sl = IoGetCurrentIrpStackLocation(irp);
 		if (irp_sl->control & SL_PENDING_RETURNED)
 			irp->pending_returned = TRUE;
 
@@ -400,13 +396,15 @@ wfastcall void WIN_FUNC(IofCompleteRequest,2)
 		 * is what we are going to call below; so we set
 		 * current_location and dev_obj for the previous
 		 * (higher) location */
+		DUMP_IRP(irp);
 		IoSkipCurrentIrpStackLocation(irp);
-
-		if (irp->current_location <= irp->stack_count)
+		if (irp->current_location < irp->stack_count)
 			dev_obj = IoGetCurrentIrpStackLocation(irp)->dev_obj;
 		else
 			dev_obj = NULL;
 
+		IOTRACE("%d, %d, %p", irp->current_location, irp->stack_count,
+			dev_obj);
 		if (irp_sl->completion_routine &&
 		    ((irp->io_status.status == STATUS_SUCCESS &&
 		       irp_sl->control & SL_INVOKE_ON_SUCCESS) ||
@@ -423,8 +421,8 @@ wfastcall void WIN_FUNC(IofCompleteRequest,2)
 				IOEXIT(return);
 		} else {
 			/* propagate pending status to next irp_sl */
-			if (irp->current_location <= irp->stack_count &&
-			    irp->pending_returned == TRUE)
+			if (irp->pending_returned &&
+			    irp->current_location < irp->stack_count)
 				IoMarkIrpPending(irp);
 		}
 	}
@@ -439,23 +437,23 @@ wfastcall void WIN_FUNC(IofCompleteRequest,2)
 		KeSetEvent(irp->user_event, prio_boost, FALSE);
 	}
 
-	IOTRACE("freeing irp %p", irp);
 	if (irp->associated_irp.system_buffer &&
 	    (irp->flags & IRP_DEALLOCATE_BUFFER))
 		ExFreePool(irp->associated_irp.system_buffer);
 	else {
+		struct mdl *mdl;
 		while ((mdl = irp->mdl)) {
 			irp->mdl = mdl->next;
 			MmUnlockPages(mdl);
 			IoFreeMdl(mdl);
 		}
 	}
+	IOTRACE("freeing irp %p", irp);
 	IoFreeIrp(irp);
 	IOEXIT(return);
 }
 
-wstdcall NTSTATUS IoPassIrpDown(struct device_object *dev_obj,
-			       struct irp *irp)
+wstdcall NTSTATUS IoPassIrpDown(struct device_object *dev_obj, struct irp *irp)
 {
 	IoSkipCurrentIrpStackLocation(irp);
 	IOEXIT(return IoCallDriver(dev_obj, irp));
@@ -471,7 +469,7 @@ wstdcall NTSTATUS IoIrpSyncComplete(struct device_object *dev_obj,
 WIN_FUNC_DECL(IoIrpSyncComplete,3)
 
 wstdcall NTSTATUS IoSyncForwardIrp(struct device_object *dev_obj,
-				  struct irp *irp)
+				   struct irp *irp)
 {
 	struct nt_event event;
 	NTSTATUS status;
@@ -909,9 +907,9 @@ wstdcall NTSTATUS WIN_FUNC(IoGetDeviceProperty,5)
 	case DevicePropertyDeviceDescription:
 	case DevicePropertyFriendlyName:
 	case DevicePropertyDriverKeyName:
-		if (wrap_is_pci_bus(wd->dev_bus_type))
+		if (wrap_is_pci_bus(wd->dev_bus))
 			RtlInitAnsiString(&ansi, "PCI");
-		else // if (wrap_is_usb_bus(wd->dev_bus_type))
+		else // if (wrap_is_usb_bus(wd->dev_bus))
 			RtlInitAnsiString(&ansi, "USB");
 		need = sizeof(wchar_t) * (ansi.max_length + 1);
 		if (buffer_len < need) {
