@@ -383,14 +383,14 @@ static struct urb *wrap_alloc_urb(struct irp *irp, unsigned int pipe,
 	USBTRACE("allocated urb: %p", urb);
 
 	urb->transfer_buffer_length = buf_len;
-	if (buf_len && buf && (
+	if (buf_len && buf && (!virt_addr_valid(buf)
 #if defined(CONFIG_HIGHMEM) || defined(CONFIG_HIGHMEM4G)
-		    1 ||
+		    || PageHighMem(virt_to_page(buf))
 #endif
-		    !virt_addr_valid(buf))) {
-		urb->transfer_buffer = usb_buffer_alloc(wd->usb.udev,
-							buf_len, alloc_flags,
-							&urb->transfer_dma);
+		    )) {
+		urb->transfer_buffer =
+			usb_buffer_alloc(wd->usb.udev, buf_len, alloc_flags,
+					 &urb->transfer_dma);
 		if (!urb->transfer_buffer) {
 			WARNING("couldn't allocate dma buf");
 			IoAcquireCancelSpinLock(&irp->cancel_irql);
@@ -664,7 +664,7 @@ static USBD_STATUS wrap_bulk_or_intr_trans(struct irp *irp)
 				 pipe_handle->bInterval);
 		USBTRACE("submitting interrupt urb %p on pipe 0x%x (ep 0x%x), "
 			 "intvl: %d", urb, urb->pipe,
-			 pipe_handle->bEndpointAddress, interval);
+			 pipe_handle->bEndpointAddress, pipe_handle->bInterval);
 	}
 	status = wrap_submit_urb(irp);
 	USBTRACE("status: %08X", status);
