@@ -133,7 +133,7 @@
 #endif
 
 #define packed __attribute__((packed))
-#define no_warn_unused __attribute((unused))
+#define no_warn_unused __attribute__((unused))
 
 typedef u8	BOOLEAN;
 typedef u8	BYTE;
@@ -279,7 +279,12 @@ struct mdl {
 	struct mdl *next;
 	CSHORT size;
 	CSHORT flags;
-	void *process;
+	/* NdisFreeBuffer doesn't pass pool, so we store pool in
+	 * unused field 'process' */
+	union {
+		void *process;
+		void *pool;
+	};
 	void *mappedsystemva;
 	void *startva;
 	ULONG bytecount;
@@ -432,7 +437,6 @@ struct nt_semaphore {
 	LONG limit;
 };
 
-//#pragma pack(push,1)
 struct nt_thread {
 	struct dispatcher_header dh;
 	/* the rest in Windows is a long structure; since this
@@ -443,7 +447,6 @@ struct nt_thread {
 	struct nt_list irps;
 	NT_SPIN_LOCK lock;
 };
-//#pragma pack(pop)
 
 #define set_dh_type(dh, type)		((dh)->type = (type))
 #define is_mutex_dh(dh)			((dh)->type == MutexObject)
@@ -804,7 +807,7 @@ enum device_usage_notification_type {
 #define METHOD_OUT_DIRECT	2
 #define METHOD_NEITHER		3
 
-#define CTL_CODE(dev_type, func, method, access)	\
+#define CTL_CODE(dev_type, func, method, access)			\
 	(((dev_type) << 16) | ((access) << 14) | ((func) << 2) | (method))
 
 #define IO_METHOD_FROM_CTL_CODE(code) (code & 0x3)
@@ -957,7 +960,7 @@ struct io_stack_location {
 #pragma pack(pop)
 #endif
 
-#define URB_FROM_IRP(irp)					\
+#define URB_FROM_IRP(irp)						\
 	(union nt_urb *)(IoGetCurrentIrpStackLocation(irp)->params.others.arg1)
 
 struct kapc {
@@ -1010,9 +1013,7 @@ struct irp {
 		LONG irp_count;
 		void *system_buffer;
 	} associated_irp;
-
 	struct nt_list threads;
-
 	struct io_status_block io_status;
 	KPROCESSOR_MODE requestor_mode;
 	BOOLEAN pending_returned;
@@ -1020,13 +1021,10 @@ struct irp {
 	CHAR current_location;
 	BOOLEAN cancel;
 	KIRQL cancel_irql;
-
 	CCHAR apc_env;
 	UCHAR alloc_flags;
-
 	struct io_status_block *user_status;
 	struct nt_event *user_event;
-
 	union {
 		struct {
 			void *user_apc_routine;
@@ -1034,10 +1032,8 @@ struct irp {
 		} async_params;
 		LARGE_INTEGER alloc_size;
 	} overlay;
-
 	void (*cancel_routine)(struct device_object *, struct irp *) wstdcall;
 	void *user_buf;
-
 	union {
 		struct {
 			union {
@@ -1077,16 +1073,16 @@ struct irp {
 #define IoGetPreviousIrpStackLocation(irp)	\
 	(IoGetCurrentIrpStackLocation(irp) + 1)
 
-#define IoSetNextIrpStackLocation(IRP)				\
+#define IoSetNextIrpStackLocation(irp)				\
 do {								\
-	(IRP)->current_location--;				\
-	IoGetCurrentIrpStackLocation(IRP)--;			\
+	(irp)->current_location--;				\
+	IoGetCurrentIrpStackLocation(irp)--;			\
 } while (0)
 
-#define IoSkipCurrentIrpStackLocation(IRP) 			\
+#define IoSkipCurrentIrpStackLocation(irp) 			\
 do {								\
-	(IRP)->current_location++;				\
-	IoGetCurrentIrpStackLocation(IRP)++;			\
+	(irp)->current_location++;				\
+	IoGetCurrentIrpStackLocation(irp)++;			\
 } while (0)
 
 static inline void
@@ -1147,7 +1143,6 @@ enum key_value_information_class {
 	KeyValuePartialInformationAlign64
 };
 
-
 struct object_attr {
 	ULONG length;
 	void *root_dir;
