@@ -31,6 +31,8 @@ static void del_stats_timer(struct wrap_ndis_device *wnd);
 static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd);
 static int ndis_remove_device(struct wrap_ndis_device *wnd);
 static void set_multicast_list(struct wrap_ndis_device *wnd);
+static int ndis_net_dev_open(struct net_device *net_dev);
+static int ndis_net_dev_close(struct net_device *net_dev);
 
 static inline int ndis_wait_comm_completion(struct wrap_ndis_device *wnd)
 {
@@ -54,7 +56,10 @@ NDIS_STATUS miniport_reset(struct wrap_ndis_device *wnd)
 
 	if (down_interruptible(&wnd->ndis_comm_mutex))
 		TRACEEXIT3(return NDIS_STATUS_FAILURE);
-	down_interruptible(&wnd->tx_ring_mutex);
+	if (down_interruptible(&wnd->tx_ring_mutex)) {
+		up(&wnd->ndis_comm_mutex);
+		TRACEEXIT3(return NDIS_STATUS_FAILURE);
+	}		
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
 	cur_lookahead = wnd->nmb->cur_lookahead;
 	max_lookahead = wnd->nmb->max_lookahead;
@@ -269,8 +274,6 @@ static NDIS_STATUS miniport_init(struct wrap_ndis_device *wnd)
 	 * fail after boot */
 	sleep_hz(HZ / 2);
 	set_bit(HW_INITIALIZED, &wnd->hw_status);
-	up(&wnd->ndis_comm_mutex);
-	up(&wnd->tx_ring_mutex);
 	hangcheck_add(wnd);
 	/* the description about NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND is
 	 * misleading/confusing; we just ignore it */
@@ -290,17 +293,12 @@ static void miniport_halt(struct wrap_ndis_device *wnd)
 	struct miniport_char *miniport;
 
 	TRACEENTER1("%p", wnd);
-	/* semaphores may already be locked, e.g., during suspend or
-	 * if device is suspended, but resume failed */
-	down_trylock(&wnd->ndis_comm_mutex);
-	down_trylock(&wnd->tx_ring_mutex);
-	if (test_bit(HW_INITIALIZED, &wnd->hw_status)) {
+	if (test_and_clear_bit(HW_INITIALIZED, &wnd->hw_status)) {
 		hangcheck_del(wnd);
 		del_stats_timer(wnd);
 		miniport = &wnd->wd->driver->ndis_driver->miniport;
 		DBGTRACE1("halt: %p", miniport->miniport_halt);
 		LIN2WIN1(miniport->miniport_halt, wnd->nmb->adapter_ctx);
-		clear_bit(HW_INITIALIZED, &wnd->hw_status);
 		/* cancel any timers left by bugyy windows driver; also free
 		 * the memory for timers */
 		while (1) {
@@ -314,6 +312,7 @@ static void miniport_halt(struct wrap_ndis_device *wnd)
 			if (!ent)
 				break;
 			wrap_timer = container_of(ent, struct wrap_timer, list);
+			wrap_timer->repeat = 0;
 			/* ktimer that this wrap_timer is associated to can't
 			 * be touched, as it may have been freed by the driver
 			 * already */
@@ -336,6 +335,7 @@ static NDIS_STATUS miniport_set_power_state(struct wrap_ndis_device *wnd,
 	DBGTRACE1("%d", state);
 	if (state == NdisDeviceStateD0) {
 		status = NDIS_STATUS_SUCCESS;
+		up(&wnd->ndis_comm_mutex);
 		if (test_and_clear_bit(HW_HALTED, &wnd->hw_status)) {
 			status = miniport_init(wnd);
 			if (status == NDIS_STATUS_SUCCESS) {
@@ -343,45 +343,33 @@ static NDIS_STATUS miniport_set_power_state(struct wrap_ndis_device *wnd,
 				set_multicast_list(wnd);
 			}
 		} else if (test_and_clear_bit(HW_SUSPENDED, &wnd->hw_status)) {
-			up(&wnd->ndis_comm_mutex);
 			status = miniport_set_int(wnd, OID_PNP_SET_POWER,
 						  state);
-			if (status == NDIS_STATUS_SUCCESS)
-				up(&wnd->tx_ring_mutex);
-			else {
-				down_interruptible(&wnd->ndis_comm_mutex);
+			if (status != NDIS_STATUS_SUCCESS)
 				WARNING("%s: setting power to state %d failed? "
 					"%08X", wnd->net_dev->name, state,
 					status);
-			}
 			if (wnd->ndis_wolopts &&
-			    wrap_is_pci_bus(wnd->wd->dev_bus_type))
+			    wrap_is_pci_bus(wnd->wd->dev_bus))
 				pci_enable_wake(wnd->wd->pci.pdev, PCI_D0, 0);
 		} else
 			return NDIS_STATUS_FAILURE;
 
 		if (status == NDIS_STATUS_SUCCESS) {
+			up(&wnd->tx_ring_mutex);
+			netif_device_attach(wnd->net_dev);
 			hangcheck_add(wnd);
 			add_stats_timer(wnd);
 			set_scan(wnd);
-			if (netif_running(wnd->net_dev)) {
-				netif_device_attach(wnd->net_dev);
-				netif_wake_queue(wnd->net_dev);
-			}
-			netif_poll_enable(wnd->net_dev);
 		} else {
 			WARNING("%s: couldn't set power to state %d; device not"
 				" resumed", wnd->net_dev->name, state);
 		}
 		TRACEEXIT1(return status);
 	} else {
-		netif_poll_disable(wnd->net_dev);
-		if (netif_running(wnd->net_dev)) {
-			netif_tx_disable(wnd->net_dev);
-			netif_device_detach(wnd->net_dev);
-		}
 		if (down_interruptible(&wnd->tx_ring_mutex))
-			WARNING("couldn't lock tx_ring_mutex");
+			TRACEEXIT1(return NDIS_STATUS_FAILURE);
+		netif_device_detach(wnd->net_dev);
 		hangcheck_del(wnd);
 		del_stats_timer(wnd);
 		status = NDIS_STATUS_NOT_SUPPORTED;
@@ -392,7 +380,7 @@ static NDIS_STATUS miniport_set_power_state(struct wrap_ndis_device *wnd,
 							  OID_PNP_ENABLE_WAKE_UP,
 							  wnd->ndis_wolopts);
 				if (status == NDIS_STATUS_SUCCESS) {
-					if (wrap_is_pci_bus(wnd->wd->dev_bus_type))
+					if (wrap_is_pci_bus(wnd->wd->dev_bus))
 						pci_enable_wake(wnd->wd->pci.pdev,
 								PCI_D0, 1);
 				} else
@@ -403,19 +391,18 @@ static NDIS_STATUS miniport_set_power_state(struct wrap_ndis_device *wnd,
 						  pm_state);
 			if (status == NDIS_STATUS_SUCCESS) {
 				set_bit(HW_SUSPENDED, &wnd->hw_status);
-				if (down_interruptible(&wnd->ndis_comm_mutex))
-					WARNING("couldn't lock ndis_comm_mutex");
 			} else
 				WARNING("suspend failed: %08X", status);
 		}
 		if (status != NDIS_STATUS_SUCCESS) {
 			WARNING("%s does not support power management; "
 				"halting the device", wnd->net_dev->name);
-			/* TODO: should we use pnp_stop_device instead? */
 			miniport_halt(wnd);
 			set_bit(HW_HALTED, &wnd->hw_status);
 			status = STATUS_SUCCESS;
 		}
+		if (down_interruptible(&wnd->ndis_comm_mutex))
+			WARNING("couldn't lock ndis_comm_mutex");
 		TRACEEXIT1(return status);
 	}
 }
@@ -444,16 +431,6 @@ static int ndis_set_mac_address(struct net_device *dev, void *p)
 		TRACEEXIT1(return -EINVAL);
 	}
 	param.type = NdisParameterString;
-	RtlInitAnsiString(&ansi, "mac_address");
-	if (RtlAnsiStringToUnicodeString(&key, &ansi, TRUE))
-		TRACEEXIT1(return -EINVAL);
-	NdisWriteConfiguration(&res, wnd->nmb, &key, &param);
-	RtlFreeUnicodeString(&key);
-	if (res != NDIS_STATUS_SUCCESS) {
-		RtlFreeUnicodeString(&param.data.string);
-		TRACEEXIT1(return -EINVAL);
-	}
-
 	RtlInitAnsiString(&ansi, "NetworkAddress");
 	if (RtlAnsiStringToUnicodeString(&key, &ansi, TRUE))
 		TRACEEXIT1(return -EINVAL);
@@ -648,7 +625,7 @@ static int miniport_tx_packets(struct wrap_ndis_device *wnd, int start, int n)
 
 static void tx_worker(void *param)
 {
-	struct wrap_ndis_device *wnd = (struct wrap_ndis_device *)param;
+	struct wrap_ndis_device *wnd = param;
 	int n;
 
 	TRACEENTER3("tx_ok %d", wnd->tx_ok);
@@ -672,6 +649,7 @@ static void tx_worker(void *param)
 			n = wnd->max_tx_packets;
 		n = miniport_tx_packets(wnd, wnd->tx_ring_start, n);
 		if (n > 0) {
+			wnd->net_dev->trans_start = jiffies;
 			wnd->tx_ring_start =
 				(wnd->tx_ring_start + n) % TX_RING_SIZE;
 			wnd->is_tx_ring_full = 0;
@@ -758,29 +736,24 @@ static int set_packet_filter(struct wrap_ndis_device *wnd, ULONG packet_filter)
 		TRACEEXIT3(return -1);
 }
 
-static int ndis_open(struct net_device *dev)
+static int ndis_net_dev_open(struct net_device *net_dev)
 {
-	ULONG packet_filter;
-	struct wrap_ndis_device *wnd = netdev_priv(dev);
+	struct wrap_ndis_device *wnd = netdev_priv(net_dev);
 
 	TRACEENTER1("%p", wnd);
-	packet_filter = NDIS_PACKET_TYPE_DIRECTED | NDIS_PACKET_TYPE_BROADCAST |
-		NDIS_PACKET_TYPE_ALL_FUNCTIONAL;
-	if (set_packet_filter(wnd, packet_filter)) {
+	if (set_packet_filter(wnd, wnd->packet_filter)) {
 		WARNING("couldn't set packet filter");
 		return -ENODEV;
 	}
-	netif_device_attach(dev);
-	netif_start_queue(dev);
+	netif_wake_queue(net_dev);
+	netif_poll_enable(net_dev);
 	return 0;
 }
 
-static int ndis_close(struct net_device *dev)
+static int ndis_net_dev_close(struct net_device *net_dev)
 {
-	if (netif_running(dev)) {
-		netif_tx_disable(dev);
-		netif_device_detach(dev);
-	}
+	netif_poll_disable(net_dev);
+	netif_tx_disable(net_dev);
 	return 0;
 }
 
@@ -795,14 +768,14 @@ static void ndis_poll_controller(struct net_device *dev)
 }
 #endif
 
-/* this function is called fom BH context */
+/* called from BH context */
 static struct net_device_stats *ndis_get_stats(struct net_device *dev)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
 	return &wnd->stats;
 }
 
-/* this function is called fom BH context */
+/* called from BH context */
 static void ndis_set_multicast_list(struct net_device *dev)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
@@ -810,7 +783,7 @@ static void ndis_set_multicast_list(struct net_device *dev)
 	schedule_wrap_work(&wnd->wrap_ndis_work);
 }
 
-/* this function is called fom BH context */
+/* called from BH context */
 struct iw_statistics *get_wireless_stats(struct net_device *dev)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
@@ -826,7 +799,7 @@ static void ndis_get_drvinfo(struct net_device *dev,
 	strncpy(info->version, DRIVER_VERSION, sizeof(info->version) - 1);
 	strncpy(info->fw_version, wnd->wd->driver->version,
 		sizeof(info->fw_version) - 1);
-	if (wrap_is_pci_bus(wnd->wd->dev_bus_type))
+	if (wrap_is_pci_bus(wnd->wd->dev_bus))
 		strncpy(info->bus_info, pci_name(wnd->wd->pci.pdev),
 			sizeof(info->bus_info) - 1);
 #ifdef CONFIG_USB
@@ -1099,26 +1072,33 @@ static void stats_timer_proc(unsigned long data)
 {
 	struct wrap_ndis_device *wnd = (struct wrap_ndis_device *)data;
 
+	TRACEENTER2("");
+	if (wnd->stats_interval <= 0)
+		TRACEEXIT2(return);
 	set_bit(COLLECT_STATS, &wnd->wrap_ndis_pending_work);
 	schedule_wrap_work(&wnd->wrap_ndis_work);
-	wnd->stats_timer.expires += 10 * HZ;
+	wnd->stats_timer.expires += wnd->stats_interval;
 	add_timer(&wnd->stats_timer);
 }
 
 static void add_stats_timer(struct wrap_ndis_device *wnd)
 {
-	init_timer(&wnd->stats_timer);
 	if (wnd->physical_medium != NdisPhysicalMediumWirelessLan)
 		return;
+	if (wnd->stats_interval < 0)
+		wnd->stats_interval *= -1;
 	wnd->stats_timer.data = (unsigned long)wnd;
 	wnd->stats_timer.function = stats_timer_proc;
-	wnd->stats_timer.expires = jiffies + 10 * HZ;
+	wnd->stats_timer.expires = jiffies + wnd->stats_interval;
 	add_timer(&wnd->stats_timer);
 }
 
 static void del_stats_timer(struct wrap_ndis_device *wnd)
 {
+	TRACEENTER2("");
+	wnd->stats_interval *= -1;
 	del_timer_sync(&wnd->stats_timer);
+	TRACEEXIT2(return);
 }
 
 static void hangcheck_proc(unsigned long data)
@@ -1127,7 +1107,9 @@ static void hangcheck_proc(unsigned long data)
 	BOOLEAN reset;
 	struct miniport_char *miniport;
 
-	TRACEENTER3("");
+	TRACEENTER2("");
+	if (wnd->hangcheck_interval <= 0)
+		TRACEEXIT2(return);
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
 	if_serialize_lock(wnd);
 	reset = LIN2WIN1(miniport->hangcheck, wnd->nmb->adapter_ctx);
@@ -1148,7 +1130,8 @@ void hangcheck_add(struct wrap_ndis_device *wnd)
 		return;
 	if (hangcheck_interval > 0)
 		wnd->hangcheck_interval = hangcheck_interval * HZ;
-	init_timer(&wnd->hangcheck_timer);
+	if (wnd->hangcheck_interval < 0)
+		wnd->hangcheck_interval *= -1;
 	wnd->hangcheck_timer.data = (unsigned long)wnd;
 	wnd->hangcheck_timer.function = hangcheck_proc;
 	wnd->hangcheck_timer.expires = jiffies + wnd->hangcheck_interval;
@@ -1158,7 +1141,11 @@ void hangcheck_add(struct wrap_ndis_device *wnd)
 
 void hangcheck_del(struct wrap_ndis_device *wnd)
 {
+	TRACEENTER2("");
+	if (wnd->hangcheck_interval > 0)
+		wnd->hangcheck_interval *= -1;
 	del_timer_sync(&wnd->hangcheck_timer);
+	TRACEEXIT2(return);
 }
 
 /* worker procedure to take care of setting/checking various states */
@@ -1189,6 +1176,7 @@ static void wrap_ndis_worker(void *param)
 NDIS_STATUS ndis_reinit(struct wrap_ndis_device *wnd)
 {
 	NDIS_STATUS status;
+
 	wnd->pm_capa = FALSE;
 	status = miniport_set_power_state(wnd, NdisDeviceStateD3);
 	if (status != NDIS_STATUS_SUCCESS) {
@@ -1614,9 +1602,11 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	DBGTRACE1("mac:" MACSTRSEP, MAC2STR(mac));
 	memcpy(&net_dev->dev_addr, mac, ETH_ALEN);
 
-	net_dev->open = ndis_open;
+	wnd->packet_filter = NDIS_PACKET_TYPE_DIRECTED |
+		NDIS_PACKET_TYPE_BROADCAST | NDIS_PACKET_TYPE_ALL_FUNCTIONAL;
+	net_dev->open = ndis_net_dev_open;
 	net_dev->hard_start_xmit = tx_skbuff;
-	net_dev->stop = ndis_close;
+	net_dev->stop = ndis_net_dev_close;
 	net_dev->get_stats = ndis_get_stats;
 	net_dev->do_ioctl = NULL;
 	if (wnd->physical_medium == NdisPhysicalMediumWirelessLan) {
@@ -1757,19 +1747,17 @@ static int ndis_remove_device(struct wrap_ndis_device *wnd)
 {
 	int tx_pending;
 
+	if (wnd->physical_medium == NdisPhysicalMediumWirelessLan)
+		miniport_set_info(wnd, OID_802_11_DISASSOCIATE, NULL, 0);
 	set_bit(SHUTDOWN, &wnd->wrap_ndis_pending_work);
 	unregister_netdevice_notifier(&netdev_notifier);
 	wnd->tx_ok = 0;
-	ndis_close(wnd->net_dev);
+	unregister_netdev(wnd->net_dev);
 	netif_carrier_off(wnd->net_dev);
 	cancel_delayed_work(&wnd->wrap_ndis_work);
-	/* In 2.4 kernels, this function is called in atomic context,
-	 * so we can't (don't need to?) wait on mutex. */
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 	/* if device is suspended, but resume failed, tx_ring_mutex is
 	 * already locked */
 	down_trylock(&wnd->tx_ring_mutex);
-#endif
 	tx_pending = wnd->tx_ring_end - wnd->tx_ring_start;
 	if (tx_pending < 0)
 		tx_pending += TX_RING_SIZE;
@@ -1785,9 +1773,7 @@ static int ndis_remove_device(struct wrap_ndis_device *wnd)
 		wnd->tx_ring_start = (wnd->tx_ring_start + 1) % TX_RING_SIZE;
 		tx_pending--;
 	}
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 	up(&wnd->tx_ring_mutex);
-#endif
 	wrap_procfs_remove_ndis_device(wnd);
 	miniport_halt(wnd);
 	ndis_exit_device(wnd);
@@ -1802,7 +1788,6 @@ static int ndis_remove_device(struct wrap_ndis_device *wnd)
 	}
 	printk(KERN_INFO "%s: device %s removed\n", DRIVER_NAME,
 	       wnd->net_dev->name);
-	unregister_netdev(wnd->net_dev);
 	free_netdev(wnd->net_dev);
 	TRACEEXIT2(return 0);
 }
@@ -1819,7 +1804,7 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	unsigned long i;
 
 	TRACEENTER2("%p, %p", drv_obj, pdo);
-	if (strlen(if_name) > (IFNAMSIZ-1)) {
+	if (strlen(if_name) >= IFNAMSIZ) {
 		ERROR("interface name '%s' is too long", if_name);
 		return STATUS_INVALID_PARAMETER;
 	}
@@ -1831,9 +1816,9 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	wd = pdo->reserved;
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 	SET_MODULE_OWNER(net_dev);
-	if (wrap_is_pci_bus(wd->dev_bus_type))
+	if (wrap_is_pci_bus(wd->dev_bus))
 		SET_NETDEV_DEV(net_dev, &wd->pci.pdev->dev);
-	if (wrap_is_usb_bus(wd->dev_bus_type))
+	if (wrap_is_usb_bus(wd->dev_bus))
 		SET_NETDEV_DEV(net_dev, &wd->usb.intf->dev);
 #endif
 	status = IoCreateDevice(drv_obj, 0, NULL,
@@ -1867,8 +1852,8 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	init_nmb_functions(nmb);
 	wnd->net_dev = net_dev;
 	wnd->ndis_irq = NULL;
-	init_MUTEX_LOCKED(&wnd->tx_ring_mutex);
-	init_MUTEX_LOCKED(&wnd->ndis_comm_mutex);
+	init_MUTEX(&wnd->tx_ring_mutex);
+	init_MUTEX(&wnd->ndis_comm_mutex);
 	init_waitqueue_head(&wnd->ndis_comm_wq);
 	wnd->ndis_comm_done = 0;
 	INIT_WORK(&wnd->tx_work, tx_worker, wnd);
@@ -1880,12 +1865,13 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	wnd->capa.encr = 0;
 	wnd->capa.auth = 0;
 	wnd->attributes = 0;
-	wnd->dma_map_count = 0;
-	wnd->dma_map_addr = NULL;
+	wnd->tx_dma_count = 0;
+	wnd->tx_dma_addr = NULL;
 	wnd->nick[0] = 0;
 	init_timer(&wnd->hangcheck_timer);
 	wnd->scan_timestamp = 0;
 	init_timer(&wnd->stats_timer);
+	wnd->stats_interval = 10 * HZ;
 	wnd->wrap_ndis_pending_work = 0;
 	memset(&wnd->essid, 0, sizeof(wnd->essid));
 	memset(&wnd->encr_info, 0, sizeof(wnd->encr_info));
@@ -1895,7 +1881,6 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	if (wd->driver->ndis_driver)
 		wd->driver->ndis_driver->miniport.shutdown = NULL;
 	wnd->stats_enabled = TRUE;
-	wnd->rx_csum.value = 0;
 	InitializeListHead(&wnd->timer_list);
 
 	fdo->reserved = wnd;
@@ -1921,6 +1906,7 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 
 int init_ndis_driver(struct driver_object *drv_obj)
 {
+	TRACEENTER1("%p", drv_obj);
 	drv_obj->drv_ext->add_device = NdisAddDevice;
 	return 0;
 }
