@@ -75,30 +75,16 @@ static void module_cleanup(void)
 	wrap_procfs_remove();
 	ndis_exit();
 	ntoskernel_exit();
-	misc_funcs_exit();
+	crt_exit();
+	rtl_exit();
 	wrapmem_exit();
 }
 
 static int __init wrapper_init(void)
 {
-	char *argv[] = {"loadndisdriver", WRAP_CMD_LOAD_DEVICES,
-#if defined(DEBUG) && DEBUG >= 1
-			"1"
-#else
-			"0"
-#endif
-			, UTILS_VERSION, NULL};
-	char *env[] = {NULL};
-	int ret;
-
 	wrapmem_init();
-	printk(KERN_INFO "%s version %s loaded (%spreempt=%s,smp=%s)\n",
+	printk(KERN_INFO "%s version %s loaded (preempt=%s,smp=%s)\n",
 	       DRIVER_NAME, DRIVER_VERSION,
-#ifdef DEBUG
-	       UTS_RELEASE ",",
-#else
-	       "",
-#endif
 #if defined CONFIG_PREEMPT
 	       "yes",
 #else
@@ -114,7 +100,7 @@ static int __init wrapper_init(void)
 #ifdef USE_OWN_WORKQUEUE
 	wrap_wq = create_singlethread_workqueue("wrap_wq");
 #endif
-	if (misc_funcs_init() || ntoskernel_init() || ndis_init()
+	if (crt_init() || rtl_init() || ntoskernel_init() || ndis_init()
 #ifdef CONFIG_USB
 	    || usb_init()
 #endif
@@ -123,17 +109,6 @@ static int __init wrapper_init(void)
 	wrap_procfs_init();
 	if (loader_init())
 		goto err;
-	DBGTRACE1("calling loadndisdriver");
-	ret = call_usermodehelper("/sbin/loadndisdriver", argv, env
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
-				  , 1
-#endif
-		);
-	if (ret) {
-		ERROR("loadndiswrapper failed (%d); check system log "
-		      "for messages from 'loadndisdriver'", ret);
-		goto err;
-	}
 	TRACEEXIT1(return 0);
 
 err:
