@@ -152,20 +152,21 @@ static int read_conf_file(char *conf_file_name, struct load_driver *driver)
 	char setting_name[MAX_SETTING_NAME_LEN];
 	char setting_value[MAX_SETTING_VALUE_LEN];
 	int ret, nr_settings;
-	int i, vendor, device, subvendor, subdevice, bus_type;
+	int i, vendor, device, subvendor, subdevice, bus;
 
 	if (lstat(conf_file_name, &statbuf)) {
-		ERROR("unable to open config file: %s", strerror(errno));
+		ERROR("unable to open config file %s: %s",
+		      conf_file_name, strerror(errno));
 		return -EINVAL;
 	}
 
 	if (sscanf(conf_file_name, "%04X:%04X.%X.conf",
-		   &vendor, &device, &bus_type) == 3) {
-		DBG("bus_type: %X", bus_type);
+		   &vendor, &device, &bus) == 3) {
+		DBG("bus: %X", bus);
 	} else if (sscanf(conf_file_name, "%04X:%04X:%04X:%04X.%X.conf",
 			  &vendor, &device, &subvendor, &subdevice,
-			  &bus_type) == 5) {
-		DBG("bus_type: %X", bus_type);
+			  &bus) == 5) {
+		DBG("bus: %X", bus);
 	} else {
 		ERROR("unable to parse conf file name %s (%d)",
 		      conf_file_name, i);
@@ -283,8 +284,7 @@ static int load_driver(int ioctl_device, char *driver_name,
 		int len;
 		struct stat statbuf;
 
-		if (strcmp(dirent->d_name, ".") == 0 ||
-		    strcmp(dirent->d_name, "..") == 0)
+		if (dirent->d_name[0] == '.')
 			continue;
 
 		if (stat(dirent->d_name, &statbuf) ||
@@ -363,180 +363,89 @@ err:
 	return -1;
 }
 
-/* check if a device is already in devices */
-static int duplicate_device(struct load_device *device, int n,
-			    struct load_device devices[])
-{
-	int i;
-
-	for (i = 0; i < n; i++)
-		if (device->vendor == devices[i].vendor &&
-		    device->device == devices[i].device &&
-		    device->subvendor == devices[i].subvendor &&
-		    device->subdevice == devices[i].subdevice)
-			return 1;
-
-	return 0;
-}
-
-/* add all devices (based on conf files) for a given driver */
-static int add_driver_devices(DIR *dir, char *driver_name, int from,
-			      struct load_device devices[])
-{
-	struct dirent *dirent;
-	int n;
-
-	n = from;
-	if (!dir || !driver_name) {
-		ERROR("invalid driver");
-		return n;
-	}
-
-	DBG("adding devices for driver %s", driver_name);
-	while ((dirent = readdir(dir))) {
-		int len;
-
-		if (n >= MAX_WRAP_DEVICES) {
-			ERROR("too many devices; increase MAX_WRAP_DEVICES "
-			      "in ndiswrapper.h and recompile or remove "
-			      "unnecessary .conf files under '%s'", confdir);
-			break;
-		}
-		if (strcmp(dirent->d_name, ".") == 0 ||
-		    strcmp(dirent->d_name, "..") == 0)
-			continue;
-
-		len = strlen(dirent->d_name);
-		if (len > 5 && strcmp(&dirent->d_name[len-5], ".conf") == 0) {
-			struct stat statbuf;
-			char *s;
-			struct load_device *device;
-
-			if (lstat(dirent->d_name, &statbuf)) {
-				ERROR("unable to open config file: %s",
-				      strerror(errno));
-				continue;
-			}
-
-			s = basename(dirent->d_name);
-			/* remove ".conf" */
-			s[strlen(s)-5] = 0;
-
-			device = &devices[n];
-			if (strlen(s) >= 11 &&
-			    sscanf(s, "%04x:%04x.%X", &device->vendor,
-				   &device->device, &device->bus_type) == 3) {
-				DBG("bus_type: %X", device->bus_type);
-				device->subvendor = DEV_ANY_ID;
-				device->subdevice = DEV_ANY_ID;
-			} else if (strlen(s) >= 21 &&
-				   sscanf(s, "%04x:%04x:%04x:%04x.%X",
-					  &device->vendor, &device->device,
-					  &device->subvendor,
-					  &device->subdevice,
-					  &device->bus_type) == 5) {
-				DBG("bus_type: %X", device->bus_type);
-			} else {
-				ERROR("file %s is not valid - ignored",
-				      dirent->d_name);
-				continue;
-			}
-			if (duplicate_device(device, n, devices))
-				DBG("device %04X:%04X is duplicate - ignored",
-				    device->vendor, device->device);
-			else {
-				strncpy(device->driver_name, driver_name,
-					sizeof(device->driver_name));
-				strncpy(device->conf_file_name, dirent->d_name,
-					sizeof(device->conf_file_name));
-				strncat(device->conf_file_name, ".conf",
-					sizeof(device->conf_file_name) -
-					strlen(device->conf_file_name));
-				DBG("device %04X:%04X:%04X:%04X is added",
-				    device->vendor, device->device,
-				    device->subvendor, device->subdevice);
-				n++;
-			}
-		}
-	}
-	DBG("number of devices in %s: %d", driver_name, n - from);
-	return n;
-}
-
-/*
- * load all installed drivers
- * returns: number of drivers loadeed successfully
- */
-static int load_all_devices(int ioctl_device)
+static int get_device(char *driver_name, int vendor, int device,
+		      int subvendor, int subdevice, int bus,
+		      struct load_device *ld)
 {
+	int ret;
 	struct stat statbuf;
+	char file[32];
+
+	DBG("%s", driver_name);
+	ret = -1;
+	if (chdir(driver_name)) {
+		DBG("couldn't chdir to %s: %s", driver_name, strerror(errno));
+		return -EINVAL;
+	}
+	if (snprintf(file, sizeof(file), "%04X:%04X:%04X:%04X.%X.conf",
+		     vendor, device, subvendor, subdevice, bus) > 0 &&
+	    stat(file, &statbuf) == 0) {
+		DBG("found %s", file);
+		ld->subvendor = subvendor;
+		ld->subdevice = subdevice;
+		ret = 0;
+	} else if (snprintf(file, sizeof(file), "%04X:%04X.%X.conf",
+			    vendor, device, bus) > 0 &&
+		   stat(file, &statbuf) == 0) {
+		DBG("found %s", file);
+		ld->subvendor = 0;
+		ld->subdevice = 0;
+		ret = 0;
+	}
+	chdir("..");
+	if (ret)
+		ld->vendor = 0;
+	else {
+		DBG("found file: %s/%s", driver_name, file);
+		ld->vendor = vendor;
+		ld->device = device;
+		ld->bus = bus;
+		strncpy(ld->driver_name, driver_name, sizeof(ld->driver_name));
+		strncpy(ld->conf_file_name, file, sizeof(ld->conf_file_name));
+	}
+	DBG("%04x, %04x, %04x, %04x", ld->vendor, ld->device, ld->subvendor,
+	    ld->subdevice);
+	return ret;
+}
+
+static int load_device(int ioctl_device, int vendor, int device,
+		       int subvendor, int subdevice, int bus)
+{
 	struct dirent  *dirent;
-	DIR *dir, *driver;
-	int loaded, res;
-	struct load_device *devices;
-	struct load_devices load_devices;
+	DIR *dir;
+	int res;
+	struct load_device load_device;
 
+	DBG("%04x, %04x, %04x, %04x", vendor, device, subvendor, subdevice);
+	memset(&load_device, 0, sizeof(load_device));
 	if (chdir(confdir)) {
+		ERROR("couldn't chdir to %s: %s", confdir, strerror(errno));
+		return -EINVAL;
+	}
+	if ((dir = opendir(".")) == NULL) {
 		ERROR("directory %s is not valid: %s",
 		      confdir, strerror(errno));
 		return -EINVAL;
 	}
-	if ((dir = opendir(confdir)) == NULL) {
-		ERROR("directory %s is not valid: %s",
-		      confdir, strerror(errno));
-		return -EINVAL;
-	}
-
-	devices = malloc(sizeof(*devices) * MAX_WRAP_DEVICES);
-	if (!devices) {
-		ERROR("couldn't allocate memory");
-		return -EINVAL;
-	}
-	loaded = 0;
 	while ((dirent = readdir(dir))) {
-		if (strcmp(dirent->d_name, ".") == 0 ||
-		    strcmp(dirent->d_name, "..") == 0)
+		DBG("%s", dirent->d_name);
+		if (dirent->d_name[0] == '.')
 			continue;
 
-		if (stat(dirent->d_name, &statbuf) ||
-		    (!S_ISDIR(statbuf.st_mode)))
-			continue;
-		if ((driver = opendir(dirent->d_name)) == NULL) {
-			ERROR("directory %s is not valid: %s",
-			      dirent->d_name, strerror(errno));
-			continue;
-		}
-		if (chdir(dirent->d_name)) {
-			ERROR("directory %s is not valid: %s",
-			      dirent->d_name, strerror(errno));
-			closedir(driver);
-			continue;
-		}
-		loaded = add_driver_devices(driver, dirent->d_name, loaded,
-					    devices);
-		chdir("..");
-		closedir(driver);
+		if (!get_device(dirent->d_name, vendor, device,
+				subvendor, subdevice, bus, &load_device))
+			break;
 	}
 	closedir(dir);
 
-	if (loaded == 0) {
-		ERROR("no valid drives found in %s; you may need to"
-		      " reinstall Windows drivers", confdir);
-		free(devices);
-		return -1;
-	}
-	load_devices.count = loaded;
-	load_devices.devices = devices;
-
-	res = ioctl(ioctl_device, WRAP_IOCTL_REGISTER_DEVICES, &load_devices);
+	DBG("%04x, %04x, %04x, %04x", load_device.vendor,
+	    load_device.device, load_device.subvendor,
+	    load_device.subdevice);
+	res = ioctl(ioctl_device, WRAP_IOCTL_LOAD_DEVICE, &load_device);
 	DBG("res: %d", res);
-	free(devices);
 
-	if (res) {
-		ERROR("couldn't load devices");
+	if (res)
 		return -1;
-	}
-	DBG("number of devices loaded: %d", loaded);
 	return 0;
 }
 
@@ -592,11 +501,6 @@ static int get_ioctl_device()
 	return fd;
 }
 
-/* two ways to call this program:
- *  first, to load all devices, use "-a" argument
- *  later, load a specific driver and device (i.e., conf file) with
- *  arguments driver name, vendor, device, subvendor, subdevice
-*/
 int main(int argc, char *argv[0])
 {
 	int i, ioctl_device, res;
@@ -640,9 +544,25 @@ int main(int argc, char *argv[0])
 		goto out;
 	}
 
-	if (strcmp(cmd, WRAP_CMD_LOAD_DEVICES) == 0) {
-		if (load_all_devices(ioctl_device))
+	if (strcmp(cmd, WRAP_CMD_LOAD_DEVICE) == 0) {
+		int vendor, device, subvendor, subdevice, bus;
+		if (argc != 9) {
+			ERROR("incorrect usage of %s (%d)", argv[0], argc);
 			res = 7;
+			goto out;
+		}
+		if (sscanf(argv[4], "%04x", &vendor) != 1 ||
+		    sscanf(argv[5], "%04x", &device) != 1 ||
+		    sscanf(argv[6], "%04x", &subvendor) != 1 ||
+		    sscanf(argv[7], "%04x", &subdevice) != 1 ||
+		    sscanf(argv[8], "%04x", &bus) != 1) {
+			ERROR("couldn't get device info");
+			res = 8;
+			goto out;
+		}
+		if (load_device(ioctl_device, vendor, device,
+				subvendor, subdevice, bus))
+			res = 9;
 		else
 			res = 0;
 	} else if (strcmp(cmd, WRAP_CMD_LOAD_DRIVER) == 0) {
